import time
import logging

from pyrasgo.api.error import APIError
from pyrasgo.config import set_session_api_key

from urllib.error import HTTPError

__all__ = [
    'connect',
    'register',
    'login'
]

def connect(api_key):
    from pyrasgo.rasgo import Rasgo
    set_session_api_key(api_key)
    return Rasgo()

def register(email: str, password: str):
    from pyrasgo.api.register import Register
    from pyrasgo.schemas.user import UserRegistration
    register = Register()
    payload = UserRegistration(
        email=email,
        password=password
    )
    try:
        response = register.login(payload)
        logging.warning('User credentials found, logging in with credentials provided. '
                        'Please use pyrasgo.login() as the preferred method in future sessions.')
        return connect(api_key=response)
    except APIError:
        logging.debug("Registering new user")

    try:
        register.register(payload=payload)
        logging.info(f"Verification Email Sent to {email}. To ensure uninterrupted access, please finish setting up your free Rasgo account by clicking the verification link in the next 24 hours")
    except APIError:
        logging.error("Unable to register user with credentials provided")
        return

    # log in new user
    try:
        time.sleep(2)
        return login(email, password)
    except APIError:
        logging.error("Unable to immediately log user in to newly registered account. " \
                "Please wait a few seconds and run the login method with " \
                "your new credentials")

def login(email: str, password: str):
    from pyrasgo.api.register import Register
    from pyrasgo.schemas.user import UserLogin
    register = Register()
    payload = UserLogin(
        email=email,
        password=password
    )
    try:
        response = register.login(payload=payload)
        return connect(api_key=response)
    except APIError:
        logging.error("Unable to log user in with credentials provided")



import time
from typing import List, Optional, Union

from snowflake.connector import ProgrammingError

from pyrasgo.utils.versioning import deprecated_without_replacement
from .error import APIError
from pyrasgo import schemas
from pyrasgo import primitives


class Update:

    def __init__(self):
        from . import Get
        from .connection import Connection
        from pyrasgo.config import get_session_api_key
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.get = Get()
        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()

    @deprecated_without_replacement('v1.0')
    def collection_attributes(self,
                              id: int,
                              attributes: List[dict]):
        """
        Create or update attributes on a Rasgo Collection

        param attributes: dict [{"key": "value"}, {"key": "value"}]
        """
        msg = 'attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]'
        if not isinstance(attributes, list):
            raise APIError(msg)
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise APIError(msg)
            for k, v in kv.items():
                attr.append(schemas.Attribute(key=k, value=v))
        attr_in = schemas.CollectionAttributeBulkCreate(collectionId = id, attributes=attr)
        return self.api._put(f"/models/{id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()

    @deprecated_without_replacement('v1.0')
    def data_source(self,
                    id: int,
                    name: Optional[str] = None,
                    domain: Optional[str] = None,
                    source_type: Optional[str] = None,
                    table: Optional[str] = None,
                    database: Optional[str] = None,
                    schema: Optional[str] = None,
                    source_code: Optional[str] = None,
                    table_status: Optional[str] = None,
                    parent_source_id: Optional[int] = None) -> primitives.DataSource:
        data_source = schemas.DataSourceUpdate(id=id,
                                           name=name,
                                           domain=domain,
                                           table=table,
                                           tableDatabase=database,
                                           tableSchema=schema,
                                           sourceCode=source_code,
                                           tableStatus=table_status,
                                           sourceType=source_type,
                                           parentId=parent_source_id)
        response = self.api._patch(f"/data-source/{id}", data_source.dict(exclude_unset=True, exclude_none=True), api_version=1).json()
        return primitives.DataSource(api_object=response)

    @deprecated_without_replacement('v1.0')
    def dataframe(self,
                  unique_id: str,
                  name: Optional[str] = None,
                  shared_status: str = None,
                  column_hash: str = None,
                  update_date: str = None) -> schemas.Dataframe:
        if shared_status not in [None, 'private', 'organization', 'public']:
            raise APIError("Valid values for shared_status are ['private', 'organization', 'public']")
        dataframe = schemas.DataframeUpdate(name=name,
                                        uniqueId=unique_id,
                                        sharedStatus=shared_status,
                                        columnHash=column_hash,
                                        updatedDate = update_date)
        response = self.api._patch(f"/dataframes/{unique_id}", dataframe.dict(exclude_unset=True, exclude_none=True), api_version=1).json()
        return schemas.Dataframe(**response)

    @deprecated_without_replacement('v1.0')
    def feature(self,
                id: int,
                display_name: Optional[str] = None,
                column_name: Optional[str] = None,
                description: Optional[str] = None,
                status: Optional[str] = None,
                tags: Optional[List[str]] = None,
                git_repo: Optional[str] = None) -> primitives.Feature:
        feature = schemas.FeatureUpdate(id=id,
                                    name=display_name,
                                    code=column_name,
                                    description=description,
                                    orchestrationStatus=status,
                                    tags=tags,
                                    gitRepo=git_repo)
        response = self.api._patch(f"/features/{id}", feature.dict(exclude_unset=True, exclude_none=True), api_version=1).json()
        return primitives.Feature(api_object=response)

    @deprecated_without_replacement('v1.0')
    def feature_attributes(self,
                           id: int,
                           attributes: List[dict]):
        """
        Create or update attributes on a feature

        param attributes: dict [{"key": "value"}, {"key": "value"}]
        """
        msg = 'attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]'
        if not isinstance(attributes, list):
            raise APIError(msg)
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise APIError(msg)
            for k, v in kv.items():
                attr.append(schemas.Attribute(key=k, value=v))
        attr_in = schemas.FeatureAttributeBulkCreate(
            featureId = id,
            attributes=attr
        )
        return self.api._put(f"/features/{id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()

    def transform(
            self,
            transform_id: int,
            name: Optional[str] = None,
            source_code: Optional[str] = None,
            type: Optional[str] = None,
            arguments: Optional[List[dict]] = None,
            description: Optional[str] = None,
            tags: Optional[Union[List[str], str]] = None
    ) -> schemas.Transform:
        """
        Updates a transform in Rasgo

        Args:
            transform_id: Id of transform to update
            name: Name of the Transform
            source_code: Source code of transform
            type: Type of transform it is. Used for categorization only
            arguments: A list of arguments to supply to the transform
                       so it can render them in the UI. Each argument
                       must be a dict with the keys: 'name', 'description', and 'type'
                       values all strings for their corresponding value
            description: Description of Transform
            tags: List of tags, or a tag (string), to set on this dataset
        """
        # Init tag array to be list of strings
        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]

        # Make request to update transform and return
        transform_update = schemas.TransformUpdate(
            name=name,
            type=type,
            description=description,
            sourceCode=source_code,
            arguments=arguments,
            tags=tags
        )
        response = self.api._put(
            f"/transform/{transform_id}",
            transform_update.dict(exclude_unset=True, exclude_none=True),
            api_version=1
        ).json()
        return schemas.Transform(**response)

    def dataset(
            self,
            dataset: primitives.Dataset,
            *,
            name: Optional[str] = None,
            description: Optional[str] = None,
            attributes: Optional[dict] = None,
    ) -> primitives.Dataset:
        """
        Update a dataset name, description, and/or attributes in Rasgo
        """
        # Raise error if trying to update a dataset in offline mode
        if not dataset._api_dataset:
            raise APIError("Can not update dataset. Needs to be saved first with `rasgo.save.dataset()`")

        dataset_update = schemas.DatasetUpdate(
            # Possible Changed Fields
            name=name,
            description=description,
            # Persist other fields in update contract so no fields set to None in update
            status=dataset._api_dataset.status,
            owner_id=dataset._api_dataset.owner_id,
            dw_table_id=dataset._api_dataset.dw_table_id,
            attributes=attributes
        )
        response = self.api._put(
            f"/datasets/{dataset._api_dataset.id}",
            dataset_update.dict(exclude_unset=True, exclude_none=True),
            api_version=2
        ).json()
        dataset_schema = schemas.Dataset(**response)
        return primitives.Dataset(
            api_dataset=dataset_schema
        )

    def dataset_table(
            self,
            dataset: primitives.Dataset,
            verbose: bool = False
    ) -> None:
        """
        Kicks off a query for re-materializing a dataset's set Dw Table.

        The dataset needs to be published first in order call this function

        Args:
            dataset: Published dataset to refresh table for
            verbose: If True will print information related to refreshing table
        """
        # We need to ensure that the API dataset is exists and is published
        if (not dataset._api_dataset or dataset.status.lower() != "published"):
            raise APIError('Can not refresh table. Dataset must first be published')

        if verbose:
            print(f"Refreshing table for dataset with id '{dataset.id}' at fqtn: '{dataset.fqtn}'")


        # Call endpoint to kick off Query to re-create table in background
        sf_query_id = self.api._put(
            f"/datasets/{dataset._api_dataset.id}/table-refresh",
            _json={'refreshInBackground': True},
            api_version=2
        ).json()
        conn = self.data_warehouse.user_connection
        while conn.is_still_running(conn.get_query_status_throw_if_error(sf_query_id)):
            if verbose:
                print("Query still executing...")
            time.sleep(1)

        if verbose:
            print(f"Done Refreshing table for dataset with id '{dataset.id}' at fqtn: '{dataset.fqtn}'")


    def column(
            self,
            dataset_column_id: int,
            display_name: Optional[str] = None,
            description: Optional[str] = None,
            attributes: Optional[dict] = None,
            tags: Optional[List[str]] = None
    ) -> schemas.DatasetColumn:
        """
        Update metadata about a dataset column

        Args:
            dataset_column_id: Dataset column id to updated. Use `dataset.columns[x].id` to retrieve
            display_name: Display name to update for this dataset column if set
            description: Description to update for this dataset column if set
            attributes: Attributes to add or update for this dataset column. Set as Key Value pairs dict
            tags: Tags to add to this dataset column

        Returns:
            Updated Dataset Column Obj
        """
        ds_col_update = schemas.DatasetColumnUpdate(
            display_name=display_name,
            description=description,
            attributes=attributes,
            tags=tags
        )
        resp = self.api._put(
            f"/dataset-columns/{dataset_column_id}",
            ds_col_update.dict(exclude_unset=True, exclude_none=True),
            api_version=2
        ).json()
        return schemas.DatasetColumn(**resp)



import json
import logging
import requests
import functools
from datetime import datetime
from time import time

from pyrasgo.config import get_session_api_key
from pyrasgo.api.connection import Connection


def track_experiment(func):
    @functools.wraps(func)
    def decorated(self, *args, **kwargs):
        try:
            api = self.api
            unique_id = self._experiment_id
        except AttributeError:
            try:
                api = Connection(api_key=get_session_api_key())
                unique_id = self._experiment_id
            except:
                logging.debug(f"Cannot create events for functions called from {self.__class__.__name__} class.")
                return func(self, *args, **kwargs)
        except KeyError:
            logging.debug(f"To create events for functions called from " \
                            f"{self.__class__.__name__} class, a dataframe must be tagged by ID.")
            return func(self, *args, **kwargs)

        if unique_id:
            event_timestamp = datetime.now().strftime("%Y-%m-%dT%H-%M-%S-%f")
            event_name = func.__name__
            event_type = self.__class__.__name__
            event_dict = {
                "event_timestamp": event_timestamp,
                "unique_id": unique_id,
                "event_type": event_type,
                "event_name": event_name,
                "event_details": {},
            }

        try:
            api._post(f"/event-logging", event_dict, api_version=1).json()
        except Exception as e:
            logging.info(f"Unable to log event with exception: {e}")
        return func(self, *args, **kwargs)
    return decorated


from .datawarehouse.base import DataWarehouse
from .datawarehouse.snowflake import SnowflakeDataWarehouse
from .dataframe.evaluate import Evaluate
from .dataframe.prune import Prune
from .dataframe.transform import Transform


from pyrasgo.primitives.collection import Collection
from pyrasgo.primitives.feature import Feature, FeatureList
from pyrasgo.primitives.dataset import Dataset
from pyrasgo.primitives.source import DataSource
from pyrasgo.primitives.accelerator import Accelerator


from typing import Optional, List, Any, Dict

from pydantic import BaseModel, Field


class AcceleratorCreate(BaseModel):
    """
    Model for creating Accelerators
    """
    # This is all we need to know. Any additional information can be read from this source.
    source_code: str = Field(alias='sourceCode')

    class Config:
        allow_population_by_field_name = True


class AcceleratorArgument(BaseModel):
    """
    Records the fact that a given Accelerator requires a given argument.

    This is meant to mimic the TransformArgument as closely as possible
    to prevent the Web UI from having to write new code
    """
    name: str
    description: str
    type: str
    extra: Optional[Dict[str, Any]]

    class Config:
        allow_population_by_field_name = True


class AcceleratorDescription(BaseModel):
    """
    Defines how to describe an accelerator
    """
    name: str
    description: Optional[str]
    author: Optional[str]
    arguments: List[AcceleratorArgument]

    class Config:
        allow_population_by_field_name = True


class Accelerator(BaseModel):
    """
    Full front-end representation of an Accelerator
    """
    id: int
    create_author: int = Field(alias="createAuthor")
    description: AcceleratorDescription
    template: str

    class Config:
        allow_population_by_field_name = True


class AcceleratorBulk(BaseModel):
    """
    Model for listing lots of Accelerators; we don't want to send
    things like the full list of arguments with each Acc when listing
    """
    id: int
    create_author: int = Field(alias="createAuthor")
    description: AcceleratorDescription

    class Config:
        allow_population_by_field_name = True


class AcceleratorArgs(BaseModel):
    """
    Represents the arguments that get passed so some given
    Accelerator to create a Dataset
    """
    name: Optional[str]
    args: Dict[str, Any]

    class Config:
        allow_population_by_field_name = True



import os
import re
from pathlib import Path
import yaml

from typing import List, Tuple
from pyrasgo.storage.datawarehouse.base import DataWarehouse
from pyrasgo.storage.datawarehouse.snowflake import SnowflakeDataWarehouse
from pyrasgo.utils.naming import random_table_name
from pyrasgo.schemas.dw_operation import Operation

DBT_MODEL_CONFIG_TEMPLATE = '''
  config(
    {config_args}
  )
'''

def operations_as_cte(
    operations: List[Operation],
):
    """
    Returns a nested CTE statement to render this op set as a CTE
    """
    # Handle single transform chains, we already have the SQL
    if len(operations) == 1:
        print(operations[0])
        return operations[0].operation_sql

    # Handle multi-transform chains
    operation_list = []
    # Need to replace old FQTNs with CTE aliases
    fqtn_mapping = {}
    for operation in operations:
        # create new aliases to replace old FQTNs with
        alias = random_table_name()
        fqtn_mapping[operation.dw_table.fqtn] = alias

        op_sql = operation.operation_sql

        # replace all instances of generated FQTNs with CTE aliases
        for fqtn, alias in fqtn_mapping.items():
            op_sql = op_sql.replace(fqtn, alias)

        # if final op, we're done. join ops and leave last one alone
        # Final op itself might be a CTE, remove the WITH and slap it on the end of this chain
        if operation == operations[-1]:
            return 'WITH ' + ', \n'.join(operation_list) + _collapse_cte(op_sql)

        operation_list.append(f'{alias} AS (\n{op_sql}\n) ')

def _collapse_cte(
    sql: str
) -> str:
    """
    Returns a collapsed CTE if the sql itself is already a CTE (starts with "with")
    """
    return re.sub(r'^(WITH)\s', ', ', sql, 1, flags=re.IGNORECASE)

def create_dbt_files(
        sql_definition: str,
        schema: List[Tuple[str, str]],
        file_name: str,
        output_directory: str = None,
        config_args: dict = None,
        include_schema: bool = False
) -> str:
    """
    Saves a dbt_project.yml and model.sql files to a directory
    """
    output_directory = output_directory or os.getcwd()
    file_name = file_name + '.sql' if file_name[-4:] != '.sql' else file_name
    filepath = os.path.join(output_directory, file_name)
    if config_args:
        # TODO: Should we validate whether args are valid dbt keywords?
        model_config = DBT_MODEL_CONFIG_TEMPLATE.format(
            config_args=config_args
        )
        model_config = '{{' + model_config + '}}'
        sql_definition = f'{model_config}\n\n{sql_definition}'
    with open(filepath, "w") as _file:
        _file.write(sql_definition)
    if include_schema:
        model_name = file_name.replace('.sql', '')
        save_schema_file(output_directory, model_name, schema, config_args)
    return filepath

def save_schema_file(
        output_directory: Path,
        model_name: str,
        schema: List[Tuple[str, str]],
        config_args: dict = None,
    ):
    """
    Writes a table def to a dbt schema file
    """
    filepath = os.path.join(output_directory, 'schema.yml')
    schema_definition = None
    columns_list = []
    for row in schema:
        columns_list.append({"name:": row[0]})
    model_dict = {"name": model_name, "columns": columns_list}
    if config_args:
        model_dict.update({"config": config_args})
    if not os.path.exists(filepath):
        schema_definition = {"version": 2, "models": [model_dict]}
    else:
        schema_definition = [model_dict]
    with open(filepath, "a") as _file:
        yaml.dump(data=schema_definition, Dumper=yaml.SafeDumper, stream=_file, sort_keys=False)
    return filepath




from typing import Dict, List, Optional

import pandas as pd

from pyrasgo.api.connection import Connection
from pyrasgo.schemas import feature as schema
from pyrasgo.schemas.attributes import Attribute, FeatureAttributes, FeatureAttributeBulkCreate
from pyrasgo.schemas.enums import Granularity


class FeatureList():
    """
        Convenience class to enable simpler console presentation,
        iteration and searching through lists of Features objects
    """

    def __init__(self, api_object):
        self.data = [Feature(api_object=entry) for entry in api_object]

    def __getitem__(self, i: int):
        return self.data[i]

    def __len__(self) -> int:
        return len(self.data)

    def __str__(self):
        ids = [str(feature.id) for feature in self]
        return (f"Features({len(self.data)} total, "
                f"ids: [{','.join(ids if len(self) < 7 else ids[0:3] + ['...'] + ids[-3:])}])")

    def __add__(self, other):
        if isinstance(other, Feature):
            return type(self)([feature.dict() for feature in self.data + [other]])
        if isinstance(other, type(self)):
            return type(self)([feature.dict() for feature in self.data + other.data])
        if isinstance(other, list) and all([isinstance(entry, Feature) for entry in other]):
            return type(self)([feature.dict() for feature in self.data + other])
        raise TypeError(f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'")

    def __repr__(self):
        return str(self)

    def filter(self, **kwargs):
        return [feature for feature in self.data
                if [feature.__getattr__(key) for key in kwargs.keys()] == list(kwargs.values())]


class Feature(Connection):
    """
    Stores a Rasgo Feature
    """

    def __init__(self, api_object, **kwargs):
        super().__init__(**kwargs)
        self._fields = schema.Feature(**api_object)

    def __getattr__(self, item):
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Feature(id={self.id}, name={self.displayName}, dataType={self.dataType}, description={self.description}, status={self.status})"

# ----------
# Properties
# ----------
    @property
    def attributes(self) -> dict:
        """
        Helper property to convert a list of dicts into a single dict
        """
        attr_dict = {}
        for a in self._fields.attributes:
            attr_dict.update({a.key: a.value})
        return attr_dict

    @property
    def dataTable(self):
        data_source = self._fields.dataSource
        if data_source:
            return data_source.dataTable
        return None

    @property
    def granularities(self):
        """
        Retrieves granularities for the feature
        """
        granularities = self._fields.granularities
        if granularities:
            return [Granularity(granularity.name.lower()) for granularity in granularities]
        return []

    @property
    def indexFields(self):
        """
        Retrieves the dimensions column name in the feature set table
        """
        return [d.columnName for d in self._fields.dimensions]

    @property
    def sourceTable(self):
        """
        Retrieves the feature set table for the feature
        """
        table_dict = self.dataTable.dict(exclude_unset=True)
        if table_dict.get('fqtn'):
            return table_dict['fqtn']
        if table_dict.get('tableName'):
            return table_dict['tableName']
        return None

    @property
    def status(self):
        """
        Returns feature status: Sandbox or Production
        """
        return 'Production' if self.orchestrationStatus == 'Productionized' else 'Sandbox'

    @property
    def tags(self) -> List[str]:
        return self._fields.tags

# -------
# Methods
# -------
    def add_attributes(self, attributes: List[dict]):
        if not isinstance(attributes, list):
            raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]')
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]')
            for k, v in kv.items():
                attr.append(Attribute(key=k, value=v))
        attr_in = FeatureAttributeBulkCreate(featureId = self.id, attributes=attr)
        self._put(f"/features/{self.id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()
        self.refresh()

    def add_tag(self, tag: str):
        if tag in self._fields.tags:
            return
        self._fields = schema.Feature(**self._put(f"/features/{self.id}/tags",
                                                api_version=1, _json={"tags": [tag, *self._fields.tags]}).json())

    def add_tags(self, tags: List[str]):
        if all(tag in self._fields.tags for tag in tags):
            return
        self._fields = schema.Feature(**self._put(f"/features/{self.id}/tags",
                                                api_version=1, _json={"tags": [*tags, *self._fields.tags]}).json())

    def build_stats(self):
        return self._post(f"/features/{self.id}/stats", api_version=1).json()

    def display_source_code(self):
        """
        Convenience function to display the sourceCode property
        """
        data_source = self._fields.dataSource
        if data_source:
            return data_source.sourceCode
        return None

    def delete_tag(self, tag: str):
        if tag not in self._fields.tags:
            return
        self._fields = schema.Feature(**self._delete(f"/features/{self.id}/tags",
                                                   api_version=1, _json={"tags": [tag]}).json())

    def delete_tags(self, tags: List[str]):
        if all(tag not in self._fields.tags for tag in tags):
            return
        self._fields = schema.Feature(**self._delete(f"/features/{self.id}/tags",
                                                   api_version=1, _json={"tags": [*tags]}).json())

    def get_stats(self):
        try:
            stats_json = self._get(f"/features/{self.id}/stats", api_version=1).json()
        except:
            return 'Cannot find stats for this feature'
        stats_obj = schema.FeatureStats(**stats_json['featureStats']) or None
        return stats_obj

    def productionize(self):
        """
        Sets a Feature's status from Sandbox to Production
        """
        print(f"Setting feature from {self.status} to Production")
        feature = schema.FeatureUpdate(id=self.id, orchestrationStatus='Productionized')
        self._fields = schema.Feature(**self._patch(f"/features/{self.id}",
                                                 api_version=1, _json=feature.dict(exclude_unset=True, exclude_none=True)).json())

    def read_into_df(self,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        """
        Pull Feature data from DataWarehouse into a pandas Dataframe
        """
        from pyrasgo.api.read import Read
        return Read().feature_data(id=self.id, filters=filters, limit=limit)

    def refresh(self):
        """
        Updates the Feature's attributes from the API
        """
        self._fields = schema.Feature(**self._get(f"/features/{self.id}", api_version=1).json())

    def rename(self, new_name: str):
        """
        Updates a Feature's display name
        """
        print(f"Renaming Feature {self.id} from {self.displayName} to {new_name}")
        feature = schema.FeatureUpdate(id=self.id, name=new_name)
        self._fields = schema.Feature(**self._patch(f"/features/{self.id}",
                                                 api_version=1, _json=feature.dict(exclude_unset=True, exclude_none=True)).json())

    def to_dict(self):
        feature_as_dict = {}
        feature_as_dict["sourceTable"] = self.sourceTable
        feature_as_dict["features"] = [{"columnName": self.columnName, "dataType": self.dataType, "displayName": self.displayName, "description": self.description}]
        feature_as_dict["dimensions"] = []
        for d in self._fields.dimensions:
            feature_as_dict["dimensions"].append({"columnName": d.columnName, "dataType": d.dataType, "granularity": d.granularity.name})
        feature_as_dict["status"] = self.orchestrationStatus
        feature_as_dict["tags"] = self.tags
        feature_as_dict["attributes"] = []
        for a in self._fields.attributes:
            feature_as_dict["attributes"].append({a.key: a.value})
        feature_as_dict["gitRepo"] = self.gitRepo
        return feature_as_dict

    def _make_table_metadata(self):
        payload = self.dataTable
        metadata = {
            "database": payload.databaseName,
            "schema": payload.schemaName,
            "table": payload.tableName,
        }
        return metadata


from typing import List, Optional

from pyrasgo.utils.versioning import deprecated_without_replacement

from .error import APIError

class Delete():

    def __init__(self):
        from .connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    @deprecated_without_replacement('v1.0')
    def collection(self, id: int) -> str:
        """
        Permanently delete a Rasgo Collection. There is no way to undo this operation.
        """
        try:
            response = self.api._delete(f"/models/{id}", api_version=1)
            if response.status_code == 200:
                return f"Collection {id} successfully deleted"
            if response.status_code == 403:
                raise APIError(f"User does not have access to delete Collection {id}")
            raise APIError(f"Problem deleting Collection {id}.")
        except:
            raise APIError(f"Problem deleting Collection {id}.")

    @deprecated_without_replacement('v1.0')
    def dataset_snapshot(self, dataset_id: int, index: int):
        return self.api._delete(f"/datasets/{dataset_id}/snapshots/{index}", api_version=2).json()

    @deprecated_without_replacement('v1.0')
    def data_source(self, id: int) -> str:
        """
        Permanently delete a Rasgo DataSource. There is no way to undo this operation.
        """
        try:
            response = self.api._delete(f"/data-source/{id}", api_version=1)
            if response.status_code == 200:
                return f"DataSource {id} successfully deleted"
            if response.status_code == 403:
                raise APIError(f"User does not have access to delete DataSource {id}")
            return f"Problem deleting DataSource {id}."
        except:
            return f"Problem deleting DataSource {id}."

    @deprecated_without_replacement('v1.0')
    def feature(self, id: int) -> str:
        """
        Permanently delete a Rasgo Feature. There is no way to undo this operation.
        """
        try:
            response = self.api._delete(f"/features/{id}", api_version=1)
            if response.status_code == 200:
                return f"Feature {id} successfully deleted"
            if response.status_code == 403:
                raise APIError(f"User does not have access to delete Feature {id}")
            return f"Problem deleting Feature {id}."
        except:
            return f"Problem deleting Feature {id}."

    def transform(self, transform_id: int) -> str:
        """
        Delete a Rasgo User Defined Transform
        """
        # NOTE: We print out error msgs on the API side
        # in the function self.api._raise_internal_api_error_if_any(response)
        # so no need to print out logic like above
        response = self.api._delete(f"/transform/{transform_id}", api_version=1)
        if response.status_code == 200:
            return f"Transform with id '{transform_id}' successfully deleted"
        if response.status_code == 403:
            raise APIError(f"User does not have access to delete Transform "
                           f"with id '{transform_id}'")
        return f"Problem deleting Transform {transform_id}."

    def dataset(self, dataset_id: int) -> str:
        """
        Delete a Dataset in Rasgo
        """
        response = self.api._delete(f"/datasets/{dataset_id}", api_version=2)
        if response.status_code == 200:
            return f"Dataset with id '{dataset_id}' successfully deleted"
        return f"Problem deleting Dataset {dataset_id}."

    def accelerator(self, accelerator_id: int) -> str:
        """
        Delete an Accelerator in Rasgo by id
        """
        response = self.api._delete(f"/accelerators/{accelerator_id}", api_version=2)
        if response.status_code == 200:
            return f"Accelerator with id '{accelerator_id}' successfully deleted"
        return f"Problem deleting Dataset {accelerator_id}."



import logging
import os
from collections import OrderedDict
from typing import Dict, Optional

import pandas as pd
import yaml
from pyrasgo.api.connection import Connection
from pyrasgo.schemas import DataType, FeaturesYML
from pyrasgo.schemas import data_source as schema
from pyrasgo.schemas.transform import Transform
from pyrasgo.utils import naming



class DataSource(Connection):
    """
    Stores a Rasgo DataSource
    """
    def __init__(self, api_object, **kwargs):
        super().__init__(**kwargs)
        self._fields = schema.DataSource(**api_object)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Source(id={self.id}, name={self.name}, sourceType={self.sourceType}, table={self.fqtn})"

    def __getattr__(self, item):
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            self.refresh()
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

# ----------
# Properties
# ----------

    @property
    def fqtn(self) -> str:
        return naming.make_fqtn(
            table=self.table,
            database=self.tableDatabase,
            schema=self.tableSchema
        )

# --------
# Methods
# --------

    def display_source_code(self):
        """
        Convenience function to display the sourceCode property
        """
        return self._fields.sourceCode

    def read_into_df(self,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        """
        Pull Source data from DataWarehouse into a pandas Dataframe
        """
        from pyrasgo.api.read import Read
        return Read().source_data(id=self.id, filters=filters, limit=limit)

    def rebuild_from_source_code(self):
        """
        Rebuild the Source using the source code
        """
        raise NotImplementedError()

    def refresh(self):
        """
        Updates the Soure's attributes from the API
        """
        self._fields = schema.DataSource(**self._get(f"/data-source/{self.id}", api_version=1).json())

    def rename(self, new_name: str):
        """
        Updates a DataSource's display name
        """
        print(f"Renaming DataSource {self.id} from {self.name} to {new_name}")
        source = schema.DataSourceUpdate(id=self.id, name=new_name)
        self._fields = schema.DataSource(**self._patch(f"/data-source/{self.id}",
                                                    api_version=1, _json=source.dict(exclude_unset=True, exclude_none=True)).json())

    def to_dict(self) -> dict:
        for f in self._fields.features:
            attr_dict = {}
            for a in f.attributes:
                attr_dict.update({a.key: a.value})
            f.attributes = attr_dict
        return FeaturesYML(
                name=self._fields.name,
                sourceTable=self._fields.table,
                dimensions=[{"columnName": d.columnName, "dataType": d.dataType, "granularity": d.granularity.name} for d in self._fields.dimensions],
                features=self._fields.features,
                sourceCode=self._fields.sourceCode,
                sourceType=self._fields.sourceType
        ).dict(exclude_unset=False, by_alias=True)

    def to_yml(self, file_name: str,
                     directory: str = None,
                     overwrite: bool = True
    ) -> str:
        if directory is None:
            directory = os.getcwd()

        if directory[-1] == "/":
            directory = directory[:-1]

        if file_name.split(".")[-1] not in ['yaml', 'yml']:
            file_name += ".yaml"

        if os.path.exists(f"{directory}/{file_name}") and overwrite:
            logging.warning(f"Overwriting existing file {file_name} in directory: {directory}")

        safe_dumper = yaml.SafeDumper
        safe_dumper.add_representer(DataType, lambda self, data: self.represent_str(str(data.value)))
        safe_dumper.add_representer(OrderedDict, lambda self, data: self.represent_mapping('tag:yaml.org,2002:map', data.items()))
        safe_dumper.ignore_aliases = lambda self, data: True

        with open(f"{directory}/{file_name}", "w") as _yaml:
            yaml.dump(data=OrderedDict(self.to_dict()), Dumper=safe_dumper, stream=_yaml)

    def _make_table_metadata(self) -> Dict:
        return {
            "database": self.dataTable.databaseName,
            "schema": self.dataTable.schemaName,
            "table": self.dataTable.tableName,
        }




from urllib import parse
import io
from typing import Dict, List, Optional, Union, TextIO, Any

from pyrasgo import primitives, schemas
from requests.exceptions import HTTPError

from pyrasgo.utils.versioning import deprecated_without_replacement

from .error import APIError

from pyrasgo.utils import polling
from ..primitives.accelerator import Accelerator
from ..schemas.accelerators import AcceleratorCreate


class Create():

    def __init__(self):
        from pyrasgo.config import get_session_api_key
        from .connection import Connection

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    @deprecated_without_replacement('v1.0')
    def collection(self, name: str,
                   type: Union[str, schemas.ModelType],
                   granularity: Union[str, schemas.Granularity],
                   description: Optional[str] = None,
                   is_shared: Optional[bool] = False) -> primitives.Collection:
        try:
            # If not enum, convert to enum first.
            model_type = type.name
        except AttributeError:
            model_type = schemas.ModelType(type)

        try:
            # If not enum, convert to enum first.
            granularity = granularity.name
        except AttributeError:
            granularity = schemas.Granularity(granularity)

        content = {"name": name,
                   "type": model_type.value,
                   "granularities": [{"name": granularity.value}],
                   "isShared": is_shared
                   }
        if description:
            content["description"] = description
        response = self.api._post("/models", _json=content, api_version=1)
        return primitives.Collection(api_object=response.json())

    def data_source(self, table: str,
                    name: str,
                    source_type: str,
                    database: Optional[str] = None,
                    schema: Optional[str] = None,
                    source_code: Optional[str] = None,
                    domain: Optional[str] = None,
                    parent_source_id: Optional[int] = None) -> primitives.DataSource:
        data_source = schemas.DataSourceCreate(name=name,
                                               table=table,
                                               tableDatabase=database,
                                               tableSchema=schema,
                                               sourceCode=source_code,
                                               domain=domain,
                                               sourceType=source_type,
                                               parentId=parent_source_id)
        response = self.api._post("/data-source", data_source.dict(exclude_unset=True), api_version=1).json()
        return primitives.DataSource(api_object=response)

    @deprecated_without_replacement('v1.0')
    def dataframe(self, unique_id: str,
                  name: str = None,
                  shared_status: str = 'organization',
                  column_hash: Optional[str] = None,
                  update_date: str = None) -> schemas.Dataframe:
        shared_status = 'organization' if shared_status not in ['public', 'private'] else shared_status
        dataframe = schemas.DataframeCreate(uniqueId=unique_id,
                                            name=name,
                                            sharedStatus=shared_status,
                                            columnHash=column_hash,
                                            updatedDate=update_date)
        try:
            response = self.api._post("/dataframes", dataframe.dict(exclude_unset=True), api_version=1).json()
        except HTTPError as e:
            error_message = f"Failed to create dataframe {unique_id}."
            if e.response.status_code == 409:
                error_message += f" This id is already in use in your organization. Dataframe IDs must be unique."
            raise APIError(error_message)
        return schemas.Dataframe(**response)

    @deprecated_without_replacement('v1.0')
    def data_source_stats(self, data_source_id: int):
        """
        Sends an api request to build stats for a specified data source.
        """
        return self.api._post(f"/data-source/profile/{data_source_id}", api_version=1).json()

    @deprecated_without_replacement('v1.0')
    def data_source_feature_stats(self, data_source_id: int):
        """
        Sends an api request to build stats for all features in a specified data source.
        """
        return self.api._post(f"/data-source/{data_source_id}/features/stats", api_version=1).json()

    @deprecated_without_replacement('v1.0')
    def feature(self,
                data_source_id: int,
                display_name: str,
                column_name: str,
                description: str,
                # data_source_column_id: int,
                status: str,
                git_repo: str,
                tags: Optional[List[str]] = None) -> primitives.Feature:
        feature = schemas.FeatureCreate(name=display_name,
                                        code=column_name,
                                        description=description,
                                        dataSourceId=data_source_id,
                                        orchestrationStatus=status,
                                        tags=tags or [],
                                        gitRepo=git_repo)
        try:
            response = self.api._post("/features/", feature.dict(exclude_unset=True), api_version=1).json()
        except HTTPError as e:
            error_message = f"Failed to create Feature {display_name}."
            if e.response.status_code == 409:
                error_message += f" {column_name} already has a feature associated with it. Try running update feature instead."
            raise APIError(error_message)
        return primitives.Feature(api_object=response)

    @deprecated_without_replacement('v1.0')
    def feature_stats(self, feature_id: int) -> Dict:
        """
        Sends an api request to build feature stats for a specified feature.
        """
        return self.api._post(f"/features/{feature_id}/stats", api_version=1).json()

    @deprecated_without_replacement('v1.0')
    def feature_importance_stats(self, id: int, payload: schemas.FeatureImportanceStats) -> Dict:
        """
        Sends an api requrest to build feature importance stats for the specified model
        """
        return self.api._post(f"/models/{id}/stats/feature-importance", payload.dict(), api_version=1).json()

    @deprecated_without_replacement('v1.0')
    def column_importance_stats(self, id: str, payload: schemas.FeatureImportanceStats) -> Dict:
        """
        Sends a json payload of importance from a dataFrame to the API so it can render in the WebApp
        """
        return self.api._post(f"/dataframes/{id}/feature-importance", payload.dict(), api_version=1).json()

    @deprecated_without_replacement('v1.0')
    def dataframe_profile(self, id: str, payload: schemas.ColumnProfiles) -> Dict:
        """
        Send a json payload of a dataframe profile so it can render in the WebApp
        """
        return self.api._post(f"/dataframes/{id}/profile", payload.dict(), api_version=1).json()

    def transform(
        self,
        *,
        name: str,
        source_code: str,
        type: Optional[str] = None,
        arguments: Optional[List[dict]] = None,
        description: Optional[str] = None,
        tags: Optional[Union[List[str], str]] = None
    ) -> schemas.Transform:
        """
        Create and return a new Transform in Rasgo
        Args:
            name: Name of the Transform
            source_code: Source code of transform
            type: Type of transform it is. Used for categorization only
            arguments: A list of arguments to supply to the transform
                       so it can render them in the UI. Each argument
                       must be a dict with the keys: 'name', 'description', and 'type'
                       values all strings for their corresponding value
            description: Description of Transform
            tags: List of tags, or a tag (string), to set on this dataset

        Returns:
            Created Transform obj
        """
        arguments = arguments if arguments else []

        # Init tag array to be list of strings
        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]

        transform = schemas.TransformCreate(
            name=name,
            type=type,
            sourceCode=source_code,
            description=description,
            tags=tags
        )
        transform.arguments = [
            schemas.TransformArgumentCreate(**x) for x in arguments
        ]
        response = self.api._post("/transform", transform.dict(), api_version=1).json()
        return schemas.Transform(**response)

    def accelerator(
        self,
        *,
        source_code: Optional[str] = None,
        source_code_path: Optional[str] = None
    ) -> Accelerator:
        """
        Given accelerator YAML source code, create an accelerator and return an obj representing it

        Supply source code as sting ( `source_code` param)  or with file path (`source_code_path` param)

        The source code can either be passed as a string or as an IO object that can yield a string.
        """
        if (not source_code and not source_code_path) or (source_code and source_code_path):
            raise ValueError("To create an Accelerator, supply the Accelerator source code as a sting with "
                             "the param `source_code` OR as a file path to your source code on you "
                             "local machine with the param `source_code_path`")

        # Load source_code file if supplied
        if source_code_path:
            fp = open(source_code_path)
            source_code= fp.read()
            fp.close()

        # Create Accelerator in API and return Accelerator Primitive
        resp = self.api._post(
            "/accelerators",
            _json= AcceleratorCreate(source_code=source_code).dict(),
            api_version=2
        ).json()
        return Accelerator(api_accelerator=schemas.Accelerator(**resp))

    def dataset_from_accelerator(
            self,
            accelerator_id: int,
            accelerator_args: Dict[str, Any],
            name: str
    ) -> None:
        """
        Applies the given set of arguments to an Accelerator to generate
        a new DRAFT Dataset in the UI, with the given name

        Args:
            accelerator_id: Id of Accelerator to apply
            accelerator_args: Arguments of the Accelerator
            name: Name of the created dataset
        """
        if not name:
            raise ValueError("Please supply a name for the Dataset to create with an Accelerator")

        # Create a Draft Dataset in the UI from the Accelerator
        apply_request = schemas.AcceleratorArgs(
            name=name,
            args=accelerator_args
        )
        resp = self.api._post(
            f"/accelerators/{accelerator_id}/apply",
            _json=apply_request.dict(),
            api_version=2
        )

        # Print message telling users Dataset creation in progress,
        # give the URL to access it, and tell them how to get the
        # PyRasgo code of the created Accelerator Dataset
        api_dataset = schemas.Dataset(**resp.json())
        draft_dataset =  primitives.Dataset(api_dataset=api_dataset)
        print(f"Draft dataset named {name!r} with id {draft_dataset.id} is being created in Rasgo.\n"
              f"View it's creation progress and UI canvas at {draft_dataset.profile()}\n\n"
              f"AFTER the dataset is finished being built in the UI, get the PyRasgo code of it's canvas using\n"
              f"    ds = rasgo.get.dataset({draft_dataset.id})\n"
              f"    print(ds.generate_py())")

    # ----------------------------------
    #  Internal/Private Create Calls
    # ----------------------------------

    def _dataset(
            self,
            *,
            name: str,
            resource_key: Optional[str] = None,
            description: Optional[str] = None,
            status: Optional[str] = 'published',
            dw_table_id: Optional[int] = None,
            dw_operation_set_id: Optional[int] = None,
            fqtn: Optional[str] = None,
            attributes: Optional[dict] = None,
            publish_ds_table_type: Optional[str] = "VIEW",
            publish_ds_table_name: Optional[str] = None,
            generate_stats: bool = True
    ) -> schemas.Dataset:
        """
        Create a Dataset in Rasgo.

        Args:
            name: Name of the dataset
            description: Description of the dataset
            status: Status of whether this datasets is published or not
            dw_table_id: DW table to associate with this Dataset
            dw_operation_set_id: Id of the Operation Set to associate with this Dataset
            fqtn: Fully qualified table name of the table to register this Dataset as
            attributes: Dictionary containing dataset attributes to be published
            publish_ds_table_type: If publishing an operations set's set DWTable after creating,
                                   select the table type for the published node ('VIEW' or 'TABLE')
            publish_ds_table_name: Optional user supplied table name to set on the
                                   published node/operation
        Returns:
            Created Dataset Obj
        """
        if (not publish_ds_table_type or publish_ds_table_type.upper() not in ("TABLE", "VIEW")):
            raise ValueError(f"table_type {publish_ds_table_type} is not usable. "
                             f"Please make sure you select either 'TABLE' or 'VIEW'.")

        dataset_create = schemas.DatasetCreate(
            name=name,
            resource_key=resource_key,
            description=description,
            status=status,
            dw_table_id=dw_table_id,
            dw_operation_set_id=dw_operation_set_id,
            attributes=attributes,
            auto_generate_stats=generate_stats
        )
        path = f"/datasets"
        if fqtn:
            path += f"?fqtn={parse.quote(fqtn)}"
        else:
            path += f"?table_type={publish_ds_table_type}"
            if publish_ds_table_name:
                encoded_table_name = parse.quote(publish_ds_table_name)
                path += f"&publish_ds_table_name={encoded_table_name}"
        response = self.api._post(
            path, dataset_create.dict(), api_version=2
        ).json()
        return schemas.Dataset(**response)

    def _operation_set_non_async(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int]
    ) -> schemas.OperationSet:
        """
        Create a operation set in Rasgo with specified operation
        and input dataset dependencies ids, in a  non-async status
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations,
            dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post(
            "/operation-sets", operation_set_create.dict(), api_version=2
        ).json()
        return schemas.OperationSet(**response)

    def _operation_set_async(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int]
    ) -> schemas.OperationSetAsyncTask:
        """
        Create a operation set in Rasgo with specified operation
        and input dataset dependencies ids
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations,
            dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post(
            "/operation-sets/async", operation_set_create.dict(), api_version=2
        ).json()
        return schemas.OperationSetAsyncTask(**response)

    def _operation_set_preview(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int]
    ) -> str:
        """
        Create a operation set in Rasgo with specified operation
        and input dataset dependencies ids
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations,
            dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post(
            "/operation-sets/offline", operation_set_create.dict(), api_version=2
        ).json()
        return response

    def _operation_set_preview_async(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int]
    ) -> schemas.OperationSetOfflineAsyncTask:
        """
        Create a operation set in Rasgo with specified operation
        and input dataset dependencies ids in an async fashion
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations,
            dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post(
            "/operation-sets/offline/async", operation_set_create.dict(), api_version=2
        ).json()
        return schemas.OperationSetOfflineAsyncTask(**response)

    def _dataset_correlation_stats(self, *, table_id: int, only_if_data_changed: Optional[bool] = True) -> None:
        """
        Trigger stats generation on a dataset
        """

        # Note: dimension_column_id should not be passed, as it has already been set at publish time
        stats_create = schemas.GenerateStat(
            dwTableId=table_id,
            dimensionColumnId=None,
            onlyIfDataChanged=only_if_data_changed
        )
        self.api._post("/stats", stats_create.dict(), api_version=2)

    def _operation_render(
            self,
            operation: schemas.OperationCreate
    ) -> str:
        """
        Test the rendering of an operation
        """
        response = self.api._post(
            "/operation/render", operation.dict(), api_version=2
        ).json()
        return response

    def _operation_set(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int],
            async_compute: bool = True,
            async_verbose: bool = False
    ) -> schemas.OperationSet:
        """
        Create and return an Operation set based on the input
        operations and dataset dependencies

        Set param `async_compute` to False to not create op with async

        Args:
            operations: List of operations to add to operation set.
                         Should be in ordered by time operation added.
            dataset_dependency_ids: Dataset ids to set as a parent for this operation set
            async_compute: Set to False not create op set in async fashion in backend/API
            async_verbose: If creating op set in async mode, set verbose to True to have verbose output

        Returns:
            Created Operation Set
        """
        if async_compute:
            from pyrasgo.api import Get

            # Submit the task request
            task_request = self._operation_set_async(
                operations=operations,
                dataset_dependency_ids=dataset_dependency_ids
            )
            operation_set_id = polling.poll_operation_set_async_status(
                task_request=task_request,
                verbose=async_verbose
            )
            return Get()._operation_set(operation_set_id)
        else:
            return self._operation_set_non_async(
                operations=operations,
                dataset_dependency_ids=dataset_dependency_ids
            )



from typing import Optional, Any, Dict, Union, List

from pyrasgo.api.connection import Connection
from pyrasgo import schemas


class Accelerator(Connection):
    """
    Primitive Class representing a Rasgo Accelerator
    """
    def __init__(
            self,
            api_accelerator: Union[schemas.Accelerator, schemas.AcceleratorBulk],
            **kwargs
    ):
        super().__init__(**kwargs)
        self._api_accelerator = api_accelerator

    def __repr__(self) -> str:
        """
        Get string representation of this Accelerator
        """
        return f"Accelerator(id={self.id}, " \
               f"name={self.name}, " \
               f"description={self.description})"

    @property
    def id(self) -> int:
        """
        Return the Id for this Accelerator
        """
        return self._api_accelerator.id

    @property
    def template(self) -> str:
        """
        Return the template for this Accelerator
        """
        # If the api object came from the bulk call,  it won't
        # have this property set. So make an api call to set it
        if not hasattr(self._api_accelerator, 'template'):
            resp = self._get(
                f"/accelerators/{self.id}", api_version=2
            ).json()
            self._api_accelerator = schemas.Accelerator(**resp)
        return self._api_accelerator.template

    @property
    def description(self) -> Optional[str]:
        """
        Return the description of this Accelerator
        """
        return self._api_accelerator.description.description

    @property
    def name(self) -> str:
        """
        Return the name of this Accelerator
        """
        return self._api_accelerator.description.name

    @property
    def arguments(self) -> List[schemas.AcceleratorArgument]:
        """
        Return the name of this Accelerator
        """
        return self._api_accelerator.description.arguments

    def apply(self, name: str, args: Dict[str, Any]) -> None:
        """
        Applies the given set of arguments to the Accelerator to generate
        a new DRAFT Dataset in the UI, with the given name.

        Args:
            name: Name of the created dataset
            args: Arguments of the Accelerator
        """
        from pyrasgo.api import Create
        Create().dataset_from_accelerator(
            accelerator_id=self.id,
            name=name,
            accelerator_args=args
        )



from datetime import datetime
from typing import Dict, List, Optional, Any

from pydantic import BaseModel, Field


class TransformArgumentCreate(BaseModel):
    name: str
    description: str
    type: str
    is_optional: Optional[bool]


class TransformArgument(TransformArgumentCreate):
    pass


class TransformCreate(BaseModel):
    name: Optional[str]
    type: Optional[str]
    sourceCode: str
    description: Optional[str]
    arguments: Optional[List[TransformArgumentCreate]]
    tags: Optional[List[str]]
    context: Optional[Dict[str, Any]]


class TransformUpdate(BaseModel):
    """
    Contract for updating a Transform
    """
    name: Optional[str]
    type: Optional[str]
    description: Optional[str]
    sourceCode: Optional[str]
    arguments: Optional[List[TransformArgumentCreate]]
    tags: Optional[List[str]]
    context: Optional[Dict[str, Any]]


class Transform(TransformCreate):
    id: Optional[int]
    arguments: Optional[List[TransformArgument]]


class TransformExecute(BaseModel):
    transform_args: Dict[str, Any] = Field(alias='transformArgs')
    transform_id: int = Field(alias='transformId')
    class Config:
        allow_population_by_field_name = True



import webbrowser
from pathlib import Path
from typing import Callable, Dict, List, Optional

import numpy as np
import pandas as pd
import yaml
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Dataset, DataSource, FeatureList
from pyrasgo.primitives.dataset import DS_PUBLISHED_STATUS
from pyrasgo.storage.dataframe import utils as df_utils
from pyrasgo.utils import ingestion, naming, udf
from pyrasgo.utils.versioning import deprecated_without_replacement

from .error import APIError, ParameterValueError


class Publish():

    def __init__(self):
        from pyrasgo.config import get_session_api_key
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        from . import Create, Get, Match, Save, Update
        from .connection import Connection

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()
        self.get = Get()
        self.match = Match()
        self.create = Create()
        self.update = Update()
        self.save = Save()

    @deprecated_without_replacement('v1.0')
    def experiment(self, df: pd.DataFrame,
                         dimensions: List[str],
                         granularity: List[str],
                         features: List[str] = None,
                         return_cli_only: bool = True,
                         verbose: bool = True) -> Collection:
        """
        Accepts a dataframe and publishes:
        - DataFrame as Rasgo DataSource
        - Dataframe columns as Rasgo Features
        - Rasgo Collection comprised of all Features

        params
        ------
        df: DataFrame: Pandas DataFrame to use as source data
        features: List[str]: List of columns to add as features.
                  If none is passed, all columns not in `dimensions` list will be registered as features
        dimensions: List[str]: List of columns to add as dimensions
        granularity: List[str]: List of granularities to apply to dimensions
        verbose: boolean: Print status statements to stdout while function executes

        Returns Rasgo Collection
        """
        features = features if features else []

        if verbose:
            print('Publishing DataFrame as a Rasgo DataSource')
        try:
            source = self.source_data(source_type='dataframe', df=df, verbose=verbose)
            if verbose:
                print('Successfully created:', source)
        except Exception as e:
            print(f'Failed to create DataSource: {e}')
            return
        if verbose:
            print('Creating Features from DataSource')
        try:
            dimensions, features = df_utils.confirm_df_columns(df, dimensions, features)
            source = self.features_from_source(source.id,
                                            features=features,
                                            dimensions=dimensions,
                                            granularity=granularity)
            features = FeatureList(**source.features)
            if verbose:
                print('Successfully created:', features)
        except Exception as e:
            print(f'Failed to create Features: {e}')
            return
        if verbose:
            print('Creating a Collection of your Features')
        try:
            collection = self.create.collection('name', 'Timeseries', granularity[0])
            collection.add_features(features)
            collection.generate_training_data()
            if verbose:
                print('Successfully created:', collection)
        except Exception as e:
            print(f'Failed to create Collection: {e}')
            return
        if not return_cli_only:
            webbrowser.open(f"app.rasgoml.com/collections/{collection.id}")
        return collection

    @deprecated_without_replacement('v1.0')
    def features(self, features_dict: Dict,
                 trigger_stats: bool = True) -> DataSource:
        f"""
        Creates or updates Features based on metadata provided in the dict

        params:
            features_dict: Valid Rasgo dict (see below)

        return:
            Rasgo DataSource

        Valid Rasgo dict format:
        -----------------------
        {ingestion.RASGO_DICT}
        """
        # TODO need to publish docs on granularity and data types
        return self.save.features_dict(features_dict, trigger_stats)

    @deprecated_without_replacement('v1.0')
    def features_from_df(self,
                         df: pd.DataFrame,
                         dimensions: List[str],
                         features: List[str],
                         granularity: List[str] = None,
                         tags: List[str] = None,
                         sandbox: bool = True):
        granularity = granularity if granularity else []
        tags = tags if tags else []
        raise NotImplementedError("`publish.features_from_df()` is a deprecated method. "
                                  "Please use `publish.features()` or `publish.features_from_source()` instead")

    @deprecated_without_replacement('v1.0')
    def features_from_source(self,
                             data_source_id: int,
                             dimensions: List[str],
                             features: List[str] = None,
                             granularity: List[str] = None,
                             tags: List[str] = None,
                             sandbox: bool = True,
                             trigger_stats: bool = True,
                             if_exists: str = 'fail',
                             verbose: bool = False) -> DataSource:
        """
        Publishes Features from an existing DataSource table

        params:
            data_source_id: ID to a Rasgo DataSource
            features: List of column names that will be features
                      If none is passed, all columns not in `dimensions` list will be registered as features
            dimensions: List of column names that will be dimensions
            granularity: List of grains that describe the form of the data. Common values are: 'day', 'week', 'month', 'second'
            tags: List of tags to be added to all features
            sandbox: boolean: False = Production, True = Sandbox
            if_exists:  fail - returns an error message if features already exists against this table
                        return - returns the features without operating on them
                        edit - edits the existing features
                        new - creates new features
            verbose: boolean: Print status statements to stdout while function executes

        return:
            Rasgo DataSource
        """
        features = features if features else []
        granularity = granularity if granularity else []
        tags = tags if tags else []

        # Check for valid DataSource
        data_source = self.get.data_source(data_source_id)
        if not data_source.table:
            raise ValueError(f"DataSource {data_source_id} is not usable. Please make sure it exists and has a valid table registered.")
        if not isinstance(dimensions, list) and all([isinstance(dimension, str) for dimension in dimensions]):
            raise TypeError('Dimensions must be provided as a list of strings.')
        if not isinstance(features, list) and all([isinstance(feature, str) for feature in features]):
            raise TypeError('Features must be provided as a list of strings.')
        if not isinstance(granularity, list):
            if isinstance(granularity, str):
                granularity = [granularity]
            else:
                raise TypeError("granularity must be provided as a list of strings")
        if not isinstance(tags, list):
            if isinstance(tags, str):
                tags = [tags]
            else:
                raise TypeError("tags must be provided as a list of strings")
        if len(granularity) > len(dimensions):
            raise APIError("Number of granularities cannot exceed number of dimensions."
                           "Dimensions are the index fields your data should join on and group by."
                           "Granularity is an attribute that describes precision of your dimensions."
                           "Consider passing more dimensions or fewer granularities.")

        # Confirm requested features and dimensions are columns in the DataSource
        columns = data_source.columns
        if len(columns) == 0:
            raise APIError("This DataSource does not have any columns.")
        dimensions, features = df_utils.confirm_list_columns([c.name for c in columns], dimensions, features)

        # Assemble put request
        put_dimensions = []
        put_features = []
        for c in columns:
            if c.name in dimensions:
                if verbose:
                    print(f'Publishing dimension {c.name}')
                # Dimension / Granularity order matching:
                try:
                    # Try to match the position of the dim with the position of the granularity
                    dim_pos = dimensions.index(c.name)
                    dim_granularity = granularity[dim_pos] if len(granularity) > 1 else granularity[0]
                except:
                    # Otherwise default to the next granularity in the list
                    dim_granularity = granularity.pop(0) if len(granularity) > 1 else granularity[0]
                put_dimensions.append(
                    api.DimensionColumnPut(
                        dataSourceColumnId = c.id,
                        columnName=c.name,
                        displayName=c.name,
                        granularityName=dim_granularity
                    )
                )

            if c.name in features:
                if verbose:
                    print(f'Publishing feature {c.name}')
                put_features.append(
                    api.FeatureColumnPut(
                        dataSourceColumnId=c.id,
                        columnName=c.name,
                        displayName=c.name,
                        description=f"Feature {c.name} created from table {data_source.table}",
                        tags=tags,
                        status="Sandboxed" if sandbox else "Productionized"
                    )
                )

        source_out = self.save.source_features(
            id = data_source.id,
            table = data_source.table,
            name = data_source.name,
            database = data_source.tableDatabase,
            schema = data_source.tableSchema,
            source_code = data_source.sourceCode,
            domain = data_source.domain,
            source_type = data_source.sourceType,
            parent_source_id= data_source.parentId,
            columns = data_source.columns,
            features = put_features,
            dimensions = put_dimensions)

        if trigger_stats:
            if verbose:
                print("Triggering stats for all features. These may take a few minutes to run.")
            self.create.data_source_feature_stats(source_out.id)
        return source_out

    @deprecated_without_replacement('v1.0')
    def features_from_source_code(self,
                                  data_source_id: int,
                                  source_code_type: str,
                                  dimensions: List[str],
                                  features: List[str] = None,
                                  granularity: List[str] = None,
                                  tags: List[str] = None,
                                  sql_definition: str = None,
                                  python_function: Callable[[pd.DataFrame], pd.DataFrame] = None,
                                  derivative_source_name: str = None,
                                  sql_view_name: str = None,
                                  sandbox: bool = True,
                                  sql_risk_override: bool = False,
                                  trigger_stats: bool = True,
                                  if_exists: str = 'fail',
                                  verbose: bool = False) -> DataSource:
        """
        Publishes a new derivative DataSource from an existing DataSource table using sql or python code
        and registers Features against columns in the new DataSource

        params:
            data_source_id: ID of an existing Rasgo DataSource that will act as the parent of the derivative source
            source_code_type: Valid string values are ['sql', 'python']
            sql_definition: Valid sql select statement that can create a view
            sql_view_name: ANSI-compliant string to name the new view
            python_function: Valid python function that accepts a dataframe as input and returns a dataframe of features
            features: List of column names that will be features
                      If none is passed, all columns not in `dimensions` list will be registered as features
            dimensions: List of column names that will be dimensions
            granularity: List of grains that describe the form of the data. Common values are: 'day', 'week', 'month', 'second'
            tags: List of tags to be added to all features
            derivative_source_name: Metadata name for the new DataSource being created
            sandbox: boolean: False = Production, True = Sandbox
            sql_risk_override: boolean: pass True if you receive a warning message about a dangerous sql keyword
            if_exists:  fail - returns an error message if features already exists against this table
                        return - returns the features without operating on them
                        edit - edits the existing features
                        new - creates new features
            verbose: boolean: Print status statements to stdout while function executes

        return:
            Rasgo DataSource
        """
        features = features if features else []
        granularity = granularity if granularity else []
        tags = tags if tags else []

        source_code_type = source_code_type.lower()
        if source_code_type not in ["sql", "python"]:
            raise ParameterValueError("source_code_type", ["sql", "python"])
        if source_code_type == "sql" and not sql_definition:
                raise ParameterValueError("sql_definition", ["Valid sql select statement that can create a veiw"])
        elif source_code_type == "python" and not python_function:
                raise ParameterValueError("python_function", ["Valid python function that accepts a dataframe as input and returns a dataframe of features"])

        # Publish the new data as a source
        # First make sure this is a valid source
        parent_source = self.get.data_source(data_source_id)
        if verbose:
            print(f"Publishing derivative source from parent source {parent_source.table}")
        child_source = self.source_data(source_type=source_code_type,
                                        sql_definition=sql_definition if source_code_type == "sql" else None,
                                        python_function=python_function if source_code_type == "python" else None,
                                        data_source_name=derivative_source_name,
                                        data_source_table_name=sql_view_name,
                                        parent_data_source_id=data_source_id,
                                        sql_risk_override=sql_risk_override,
                                        trigger_stats=trigger_stats,
                                        if_exists=if_exists)

        if not child_source:
            raise APIError(f"Unable to create a derivative source from the {source_code_type} code provided."
                            "Please check that it creates a valid view or dataframe and try again.")

        # Then publish features
        if verbose:
            print(f"Publishing features from derivative source {child_source.id}")
        return self.features_from_source(data_source_id=child_source.id,
                                         features=features,
                                         dimensions=dimensions,
                                         granularity=granularity,
                                         tags=tags,
                                         sandbox=sandbox,
                                         trigger_stats=trigger_stats,
                                         if_exists=if_exists)

    @deprecated_without_replacement('v1.0')
    def features_from_yml(self,
                          yml_file: str,
                          sandbox: Optional[bool] = True,
                          git_repo: Optional[str] = None,
                          trigger_stats: bool = True,
                          verbose: bool = False) -> DataSource:
        """
        Publishes metadata about Features to Pyrasgo

        params:
            yml_file: Rasgo compliant yml file that describes the feature(s) being created
            sandbox: Status of the features (True = 'Sandboxed' | False = 'Productionized')
            git_repo: Filepath string to these feature recipes in git
            verbose: boolean: Print status statements to stdout while function executes

        return:
            Rasgo DataSource
        """
        with open(yml_file) as fobj:
            if verbose:
                print("Unpacking yml")
            contents = yaml.load(fobj, Loader=yaml.SafeLoader)
        if isinstance(contents, list):
            raise APIError("More than one feature set found, please pass in only one feature set per yml")
        else:
            if verbose:
                print("Publishing features")
            features = self.save.features_dict(contents, trigger_stats)
        return features

    @deprecated_without_replacement('v1.0')
    def source_data(self,
                    source_type: str,
                    file_path: Path = None,
                    df: pd.DataFrame = None,
                    table: str = None,
                    table_database: Optional[str] = None,
                    table_schema: Optional[str] = None,
                    sql_definition: str = None,
                    python_function: Optional[Callable[[pd.DataFrame], pd.DataFrame]] = None,
                    data_source_name: str = None,
                    data_source_domain: str = None,
                    data_source_table_name: str = None,
                    parent_data_source_id: Optional[int] = None,
                    sql_risk_override: bool = False,
                    trigger_stats: bool = True,
                    if_exists: str = 'fail',
                    verbose: bool = True
                    ) -> DataSource:
        """
        Push a csv, Dataframe, or table to a Snowflake table and register it as a Rasgo DataSource (TM)

        NOTES: csv files will import all columns as strings

        params:
            source_type: Values: ['csv', 'dataframe', 'table', 'sql', 'python']
            df: pandas DataFrame (only use when passing source_type = 'dataframe')
            file_path: full path to a file on your local machine (only use when passing source_type = 'csv')
            table: name of a valid Snowflake table in your Rasgo account (only use when passing source_type = 'table')
            table_database: Optional: name of the database of the table passed in 'table' param (only use when passing source_type = 'table')
            table_schema: Optional: name of the schema of the table passed in 'table' param (only use when passing source_type = 'table')
            sql_definition: Optional: valid SQL string that will create a new view in your DataWarehous
            python_function: Optional: valid python function that accepts a dataframe as input and returns a dataframe of features
            data_source_name: Optional name for the DataSource (if not provided a random string will be used)
            data_source_table_name: Optional name for the DataSource table in Snowflake (if not provided a random string will be used)
            data_source_domain: Optional domain for the DataSource (default is NULL)
            parent_data_source_id: Optional ID of a valid Rasgo DataSource that is a parent to this DataSource (default is NULL)
            sql_risk_override: boolean: pass True if you receive a warning message about a dangerous sql keyword
            if_exists: Values: ['fail', 'append', 'replace'] directs the function what to do if a DataSource already exists with this table name  (defaults to fail)
            verbose: boolean: Print status statements to stdout while function executes

        return:
            Rasgo DataSource
        """
        # V1 Trade-offs / Possible Future Enhancements
        # --------------
        # csv's upload with all columns as string data type
        # uploading csv locally vs. calling the post/data-source/csv endpoint

        # Validate inputs
        vals = ["csv", "dataframe", "table", "sql", "python"]
        source_type = source_type.lower()
        _org_defaults = self.api._get_default_namespace()

        if source_type not in vals:
            raise ParameterValueError("source_type", vals)
        if source_type == "csv" and not Path(file_path).exists():
                raise FileNotFoundError("Please pass in a valid file path using the `file_path` parameter")
        if source_type == "dataframe":
            if not isinstance(df, pd.DataFrame) or df.empty:
                raise ValueError("Please pass in a valid DataFrame using the `df` parameter")
        if source_type == 'sql':
            if not sql_definition:
                raise ValueError('Please pass in a valid SQL string using the `sql_definition` parameter')
            #check for dangerous sql
            if naming._is_scary_sql(sql_definition) and not sql_risk_override:
                # NOTE: In the future we will want to send an email to the account owner
                #       or require special permissions to run with a sql_risk_override param
                raise ValueError("It looks like the sql string you passed may have some potentially dangerous keywords in it like DELETE or DROP or TRUNCATE. "
                                 "This SQL will be used to create a view in your DataWarehouse. Please ensure this is the code you want to run. "
                                 "If you're confident in your code, please pass in the parameter `sql_risk_override=True` to run this code.")
            if not naming._is_valid_view_sql(sql_definition) and not sql_risk_override:
                raise ValueError("It looks like the sql string you passed is missing basic elements like a SELECT or FROM clause. "
                                 "This SQL will be used to create a view in your DataWarehouse. Please ensure this is the code you want to run. "
                                 "If you're confident in your code, please pass in the parameter `sql_risk_override=True` to run this code.")
        if source_type == "table":
            if not table:
                raise ValueError("Please pass in a valid table using the `table` parameter")
            if if_exists in ["append", "replace"]:
                # Raise this as an error after we've confirmed there will be no backwards compat issues
                print("Data Sources built from SQL tables do not physically append or replace data, proceeding with a no-op registration.")
            parsed_database, parsed_schema, parsed_table = naming.parse_fqtn(table, _org_defaults)
            table = parsed_table
            table_database = table_database or parsed_database
            table_schema = table_schema or parsed_schema
            try:
                src_table = self.data_warehouse.get_source_table(table_name=table, database=table_database, schema=table_schema, record_limit=10)
                if src_table.empty:
                    raise APIError(f"Source table {table} is empty or this role does not have access to it.")
            except:
                raise APIError(f"Source table {table} does not exist or this role does not have access to it.")
        if data_source_table_name == "":
            raise ValueError("The `data_source_table_name` parameter was passed as a blank string, please pass in a valid SQL table name")

        if source_type == "python":
            if not python_function:
                raise ValueError("Please pass in a valid python function using the `python_function` parameter")
            if not parent_data_source_id:
                raise ValueError("Please pass in a valid data source for processing via python function ")
            udf.is_valid_udf(python_function)

        # Determine the ouput table based on user inputs
        if source_type in ["csv", "dataframe", "sql", "python"]:
            table_name = data_source_table_name or naming.random_table_name()
        else: #source_type == "table":
            table_name = table
        table_name = table_name.upper()
        table_database = table_database if source_type == "table" else _org_defaults.get('database')
        table_schema = table_schema if source_type == "table" else _org_defaults.get('schema')
        table_fqtn = naming.make_fqtn(table=table_name, database=table_database, schema=table_schema)

        # Check if a DataSource already exists
        data_source = self.match.data_source(fqtn=table_fqtn)
        if data_source:
            # If it does, override to the existing table
            table_name = data_source.table
            table_database = data_source.tableDatabase
            table_schema = data_source.tableSchema
            table_fqtn = naming.make_fqtn(table=table_name, database=table_database, schema=table_schema)

            # Then handle input directives
            vals = ["append", "fail", "replace"]
            msg = f"DataSource {data_source.id} already exists. "
            if data_source.organizationId != self.api._profile.get('organizationId'):
                raise APIError(msg+"This API key does not have permission to replace it.")
            if data_source.sourceType.lower() != source_type:
                raise APIError(msg+f"Your input parameters would edit the source_type from {data_source.sourceType.lower()} to {source_type}. "
                                    "To change DataSource attributes, use the `update.data_source` method. "
                                    "To update this source table, ensure your input parameters match the DataSource definition.")
            if if_exists not in vals:
                raise ParameterValueError("if_exists", vals)
            if if_exists == 'fail':
                raise APIError(msg+"Pass if_exists='replace' or 'append' to proceed.")

        # Determine operation to perform: [create, append, replace, register, no op]
        _operation = self._source_table_operation(source_type, if_exists, table_fqtn, org_defaults=_org_defaults)
        if verbose:
            print(f"Proceeding with operation {_operation}")

        # Upload to Snowflake
        # Path: csv & dataframe
        if source_type in ["csv", "dataframe"]:
            if source_type == "csv":
                df = pd.read_csv(file_path)
            if verbose:
                print("Uploading to table:", table_name)
            df_utils._cleanse_sql_dataframe(df)
            self.data_warehouse.write_dataframe_to_table(df, table_name=table_name, append=True if if_exists=="append" else False)
            self.data_warehouse.grant_table_ownership(table=table_name, role=self.data_warehouse.publisher_role)
            self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role)

        # Path: python
        if source_type == "python":
            # Get df from parent source, run function, upload result df as new table
            parent_table_name = self.get.data_source(parent_data_source_id).table
            source_df = self.data_warehouse.get_source_table(table_name=parent_table_name)
            result_df = python_function(source_df)
            if verbose:
                print("Creating table with python transformed dataframe:", table_name)
            df_utils._cleanse_sql_dataframe(result_df)
            self.data_warehouse.write_dataframe_to_table(result_df, table_name=table_name, append=True if if_exists=="append" else False)
            self.data_warehouse.grant_table_ownership(table=table_name, role=self.data_warehouse.publisher_role)
            self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role)

        # Path: sql
        if source_type == "sql":
            view_query = f"CREATE OR REPLACE VIEW {table_name} AS {sql_definition};"
            if verbose:
                print("Creating sql view as:", view_query)
            self.data_warehouse.execute_query(view_query)
            self.data_warehouse.grant_table_ownership(table=table_name, role=self.data_warehouse.publisher_role)
            self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role)

        # Path: table
        if source_type == "table":
            if verbose:
                print(f"Granting read access on table {table_name} to {self.data_warehouse.reader_role.upper()}")
            # NOTE: grant_table_acess is intentional here
            # In other methods, we create a table with the rasgo user role and want to hand if off to the reader role
            # In this case, the table is likely part of a pre-existing rbac model and we just want to grant rasgo access
            self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role, database=table_database, schema=table_schema)

        source_code = None
        if source_type == 'sql':
            source_code = sql_definition
        elif source_type == 'python':
            source_code = udf.stringify_udf(python_function)

        # Publish DataSource
        data_source = self.save.data_source(name=data_source_name or table_name,
                                            table=table_name,
                                            database=table_database if source_type == 'table' else _org_defaults.get('database'),
                                            schema=table_schema if source_type == 'table' else _org_defaults.get('schema'),
                                            source_code=source_code,
                                            domain=data_source_domain,
                                            source_type=source_type,
                                            parent_source_id=parent_data_source_id,
                                            if_exists='edit')
        if data_source:
            if trigger_stats:
                self.create.data_source_stats(data_source.id)
            return data_source
        else:
            raise APIError("DataSource failed to upload")

    @deprecated_without_replacement('v1.0')
    def _source_table_operation(self,
                                source_type: str,
                                if_exists: str,
                                to_fqtn: str,
                                from_fqtn: Optional[str] = None,
                                org_defaults: Optional[Dict] = None):
        """
        Called by publish_source_data:
            Given a source_type and tables, return the operation that should be performed to publish this table
        """
        to_database, to_schema, to_table = naming.parse_fqtn(to_fqtn, org_defaults)
        data_source_exists = True if self.match.data_source(fqtn=to_fqtn) is not None else False

        try:
            dest_table = self.data_warehouse.get_source_table(table_name=to_table, database=to_database, schema=to_schema, record_limit=10)
            is_dest_table_empty = dest_table.empty
        except:
            is_dest_table_empty = True

        if source_type in ["csv", "dataframe", "sql", "python"]:
            if not data_source_exists:
                if is_dest_table_empty:
                    return "create"
                else: #not is_dest_table_empty
                    raise APIError(f"A table named {to_fqtn} already exists, but is not registered as a Rasgo DataSource. "
                                   f"Try running this function with params: source_type='table', table='{to_table}'. "
                                    "If this wasn't an intentional match, run this function again to generate a new table name.")
            elif data_source_exists:
                if is_dest_table_empty:
                    return "create"
                else: #not is_dest_table_empty
                    return if_exists
            else:
                raise APIError("Could not determine what operation to perform.")
        elif source_type == "table":
            if not data_source_exists:
                return "register"
            elif not is_dest_table_empty and if_exists in ["append", "replace"]:
                print(f"pyRasgo does not support {if_exists} operations on tables yet.")
            return "no op"
        else:
            raise APIError("Could not determine what operation to perform.")

    def table(
            self,
            fqtn: str,
            name: Optional[str] = None,
            description: Optional[str] = None,
            parents: Optional[List[Dataset]] = None,
            verbose: bool = False,
            attributes: Optional[dict] = None,
            generate_stats: bool = True
    ) -> Dataset:
        """
        Register an existing table as a Rasgo Dataset

        params:
            fqtn: The fully qualified table name of the table to register
            name: Optional name to apply to this Rasgo Dataset
            description: Optional description for this Rasgo Dataset
            parents: Set Parent Dataset dependencies for this table dataset. Input as list of dataset primitive objs.
            verbose: Print status statements to stdout while function executes if true
            attributes: Dictionary with metadata about the Dataset
            generate_stats: If True (default) will generate stats for table dataset when published
        return:
            Rasgo Dataset
        """
        # Validate all parent ds Ids exist if passed
        # Calling rasgo.get.dataset(<id>) will raise error if doesn't
        parents = parents if parents else []
        parent_ids = [ds.id for ds in parents]
        for p_ds_id in parent_ids:
            self.get.dataset(p_ds_id)

        if verbose:
            print("Publishing table as Rasgo dataset")

        if fqtn.count(".") != 2:
            raise ValueError(f"'{fqtn}' is not a valid fully qualified table name. "
                             f"FQTNs should follow the format DATABASE.SCHEMA.TABLE_NAME.  "
                             f"Please pass a valid FQTN and try again")

        table_database, table_schema, table_name = fqtn.split('.')

        try:
            src_table = self.data_warehouse.get_source_table(
                table_name=table_name,
                database=table_database,
                schema=table_schema,
                record_limit=10
            )
            if src_table.empty:
                raise APIError(f"Source table {table_name} is empty or this role "
                               f"does not have access to it.")
        except:
            raise APIError(f"Source table {table_name} does not exist or this role "
                           f"does not have access to it.")

        # TODO: Re-implement when search by fqtn is available. For now, we'll trust the API to enforce uniqeness on fqtns
        # Check if a Dataset already exists
        # dataset = self.match.dataset(fqtn=fqtn)
        # if dataset:
        #     raise APIError(f"Dataset {dataset.id} already exists in Rasgo")

        # NOTE: grant_table_access is intentional here
        # In other methods, we create a table with the rasgo user role and want to hand if off to the reader role
        # In this case, the table is likely part of a pre-existing rbac model and we just want to grant rasgo access
        self.data_warehouse.grant_table_access(
            table=table_name,
            role=self.data_warehouse.reader_role,
            database=table_database,
            schema=table_schema
        )

        # Create operation set with parent dependencies
        # set for this dataset
        operation_set = self.create._operation_set(
            operations=[],
            dataset_dependency_ids=parent_ids,
            async_compute=False
        )

        # Publish Dataset with operation set created above
        dataset = self.create._dataset(
            name=name or table_name,
            description=description,
            fqtn=fqtn,
            status=DS_PUBLISHED_STATUS,
            attributes=attributes,
            dw_operation_set_id=operation_set.id,
            generate_stats=generate_stats
        )
        # Raise API error if backend error creating dataset
        if not dataset:
            raise APIError("DataSource failed to upload")

        # Return dataset if no error
        if verbose:
            print("Done publishing table as Rasgo dataset")
        return Dataset(api_dataset=dataset)

    def df(
            self,
            df: pd.DataFrame = None,
            name: Optional[str] = None,
            description: Optional[str] = None,
            dataset_table_name: str = None,
            parents: Optional[List[Dataset]] = None,
            verbose: bool = False,
            attributes: Optional[dict] = None,
            fqtn: Optional[str] = None,
            if_exists: Optional[str] = None,
            generate_stats: bool = True
    ) -> Dataset:
        """
        Push a Pandas Dataframe a Data Warehouse table and register it as a Rasgo Dataset

        params:
            df: pandas DataFrame
            name: Optional name for the Dataset (if not provided a random string will be used)
            description: Optional description for this Rasgo Dataset
            dataset_table_name: Optional name for the Dataset table in Snowflake (if not provided a random string will be used)
            parents: Set Parent Dataset dependencies for this df dataset. Input as list of dataset primitive objs.
            verbose: Print status statements to stdout while function executes if true
            attributes: Dictionary with metadata about the Dataset
            fqtn: If appending to an existing table via DataFrame, include the Fully Qualified Table Name here
            if_exists: Values: ['fail', 'append', 'overwrite'] directs the function what to do if a FTQN is passed, and represents an existing Dataset
            generate_stats: If True (default) will generate stats for df dataset when published
        return:
            Rasgo Dataset
        """
        # Make sure no incompatible dw dtypes in df uploading
        _raise_error_if_bad_df_dtypes(df)

        # Validate all parent ds Ids exist if passed
        # Calling rasgo.get.dataset(<id>) will raise error if doesn't
        parents = parents if parents else []
        parent_ids = [ds.id for ds in parents]
        for p_ds_id in parent_ids:
            self.get.dataset(p_ds_id)

        if_exists_vals = ["overwrite", "append", "fail"]

        if (if_exists and if_exists.lower() not in if_exists_vals):
            raise ParameterValueError("if_exists", if_exists_vals)

        if (fqtn and fqtn.count(".") != 2):
            raise ValueError(
                f"'{fqtn}' is not a valid fully qualified table name. "
                f"FQTNs should follow the format DATABASE.SCHEMA.TABLE_NAME.  "
                f"Please pass a valid FQTN and try again"
            )

        if (if_exists and not fqtn):
            raise ValueError(
                f"`if_exists` passed as '{if_exists}', but no FQTN was passed. "
                "In order to amend an existing Dataset, please pass the FQTN "
                "of that dataset"
            )
            pass

        if verbose:
            print("Publishing df as Rasgo dataset")

        if fqtn:
            db_name, schema_name, table_name = fqtn.split(".")
            # Get dataset matching FQTN
            ds = self.get.dataset(
                fqtn=fqtn
            )
            if verbose:
                print(f"Found Dataset {ds.id} matching FQTN {fqtn}")

            # If no Dataset matching the FQTN, fail and warn
            if not ds:
                raise ValueError(f"Dataset with FQTN {fqtn} not found. Please confirm FQTN.")
            else:
                if (if_exists and if_exists.lower() == "overwrite"):
                    # Users should only be able to overwrite datasets in their own organization
                    if ds._api_dataset.organization_id != self.api._profile.get('organizationId'):
                        raise APIError(f"Dataset {ds.id} already exists. This API key does not have permission to replace it.")
                    self.data_warehouse.write_dataframe_to_table(df, table_name=table_name, append=False)
                    if verbose:
                        print(f"Dataset {ds.id} with FQTN {fqtn} successfully overwritten")
                elif (if_exists and if_exists.lower() == "append"):
                    self.data_warehouse.write_dataframe_to_table(df, table_name=table_name, append=True)
                    if verbose:
                        print(f"Data successfully appended to Dataset {ds.id} with FQTN {fqtn}")
                elif (if_exists and if_exists.lower() == "fail"):
                    raise ValueError(f"FQTN {fqtn} already exists, and {if_exists} was passed for `if_exists`. Please confirm the FQTN or choose another value for `if_exists`")

                self.data_warehouse.grant_table_ownership(table=table_name, role=self.data_warehouse.publisher_role, database=db_name, schema=schema_name)
                self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role, database=db_name, schema=schema_name)

            # return the dataset we have
            return ds
        else:
            _org_defaults = self.api._get_default_namespace()
            table_name = (dataset_table_name or naming.random_table_name()).upper()
            name = name or table_name
            table_database = _org_defaults.get('database')
            table_schema = _org_defaults.get('schema')
            naming_fqtn = naming.make_fqtn(table=table_name, database=table_database, schema=table_schema)

            df_utils._cleanse_sql_dataframe(df)
            self.data_warehouse.write_dataframe_to_table(df, table_name=table_name, append=False)
            self.data_warehouse.grant_table_ownership(table=table_name, role=self.data_warehouse.publisher_role)
            self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role)

            # Create dataset based on new table FQTN created from df
            dataset = self.table(
                fqtn=naming_fqtn,
                name=name,
                description=description,
                verbose=verbose,
                attributes=attributes,
                parents=parents,
                generate_stats=generate_stats
            )
        if verbose:
            print("Done publishing df as Rasgo dataset")
        return dataset

    def dataset(
            self,
            dataset: Dataset,
            name: str,
            resource_key: Optional[str] = None,
            description: Optional[str] = None,
            verbose: bool = False,
            time_index: str = None,
            attributes: Optional[dict] = None,
            table_type: Optional[str] = "VIEW",
            table_name: Optional[str] = None,
            generate_stats: bool = True
    ) -> Dataset:
        """
        Saves a transformed Dataset in Rasgo to published
        Args:
            dataset: Dataset to save
            name: Name of dataset
            resource_key: A table-safe key used to identify this dataset
            description: Description of dataset
            verbose: If true will print save progress status
            time_index: If the dataset is a time-series with a date column, pass the name of the date column here
            attributes: Dictionary with metadata about the Dataset
            table_type: Type of object to create in snowflake. Can be "TABLE" or "VIEW"
            table_name: Data Warehouse Table Name to set for this DS's published operation
            generate_stats: If True (default) will generate stats for dataset when published
        """
        # Saving of previously-existing Datasets is not allowed
        if dataset._api_dataset:
            raise APIError(f"This Dataset already exists in Rasgo. "
                           f"{dataset}. Transform the dataset to save it.")
        if verbose:
            print(f"Saving Dataset with name={name!r} description={description!r} resource_key={resource_key}...")
        operation_set = dataset._get_or_create_op_set()
        if time_index:
            if (attributes and "time_index" not in attributes):
                attributes["time_index"] = time_index
            elif (attributes and "time_index" in attributes):
                raise ValueError(
                    f"Timeseries index explicitly defined as"
                    f" {time_index} in parameters, but defined as "
                    f"{attributes['time_index']} in attributes dict. "
                    f"Please choose to publish either in attributes or "
                    f"as a seperate parameter, but not both."
                )
            else:
                attributes = {"time_index": time_index}

        dataset = self.create._dataset(
            name=name,
            resource_key=resource_key,
            description=description,
            status=DS_PUBLISHED_STATUS,
            dw_table_id=operation_set.operations[-1].dw_table_id,
            dw_operation_set_id=operation_set.id,
            attributes=attributes,
            publish_ds_table_type=table_type,
            publish_ds_table_name=table_name,
            generate_stats=generate_stats
        )
        dataset = Dataset(api_dataset=dataset, api_operation_set=operation_set)
        if verbose:
            print(f"Finished Saving {dataset}")
        return dataset


# ------------------------------------------
#  Private Helper Funcs for Publish Class
# ------------------------------------------

def _raise_error_if_bad_df_dtypes(df: pd.DataFrame) -> None:
    """
    Raise an API error is any dtypes in the pandas dataframe,
    which are being pushed to the data warehouse aren't compatible.

    Raise proper error message if so
    """
    for col_name in df:
        col = df[col_name]
        if col.dtype.type == np.datetime64:
            raise APIError(f"Can't publish pandas Df to Rasgo. Df column "
                           f"'{col_name}' needs to be converted to proper datetime format.\n\n"
                           f"If your column is a **DATE** use `pd.to_datetime(df[<col_name>]).dt.date` to convert it\n"
                           f"If your column is a **TIMESTAMP** use `pd.to_datetime(final_df['col_name']).dt.tz_localize('UTC')`")
        else:
            pass



from typing import List, Optional

from pyrasgo import schemas
from pyrasgo.primitives import Dataset, DataSource, Feature
from pyrasgo.primitives.dataset import DS_PUBLISHED_STATUS
from pyrasgo.utils.versioning import deprecated_without_replacement
from pyrasgo.utils import ingestion, naming
from tqdm import tqdm

from .error import APIError, APIWarning


class Save():

    def __init__(self):
        from pyrasgo.config import get_session_api_key
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        from . import Create, Get, Match, Update
        from .connection import Connection

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()
        self.get = Get()
        self.match = Match()
        self.create = Create()
        self.update = Update()

    @deprecated_without_replacement('v1.0')
    def data_source(self, table: str,
                    name: str = None,
                    database: Optional[str] = None,
                    schema: Optional[str] = None,
                    source_code: Optional[str] = None,
                    domain: Optional[str] = None,
                    source_type: Optional[str] = None,
                    parent_source_id: Optional[int] = None,
                    if_exists: str = 'return'
                    ) -> DataSource:
        """
        Creates or updates a DataSource depending on the defined parameters

        if_exists:  return - returns the source without operating on it
                    edit - edits the existing source
                    new - creates a new source
        """
        fqtn = naming.make_fqtn(table=table, database=database, schema=schema)
        ds = self.match.data_source(fqtn=fqtn)
        if ds:
            if if_exists == 'return':
                return ds
            elif if_exists in ['edit', 'replace', 'append']:
                return self.update.data_source(id=ds.id, name=name, table=table, database=database, schema=schema, source_code=source_code, domain=domain, source_type=source_type, parent_source_id=parent_source_id)
            else:
                raise APIError(f"Table {fqtn} is already registered as Data Source {ds.id}")
        return self.create.data_source(name=name, table=table, database=database, schema=schema, source_code=source_code, domain=domain, source_type=source_type, parent_source_id=parent_source_id)

    @deprecated_without_replacement('v1.0')
    def dataframe(self, unique_id: str,
                  name: str = None,
                  shared_status: str = None,
                  column_hash: Optional[str] = None,
                  update_date: str = None,
                  if_exists: str = 'return') -> schemas.Dataframe:
        """
        Creates or returns a Dataframe depending on of the defined parameters

        if_exists:  return - returns the source without operating on it
                    edit - edits the existing source
                    new - creates a new source
        """
        df = self.match.dataframe(unique_id=unique_id)
        if df:
            if if_exists == 'return':
                return df
            elif if_exists in ['edit', 'replace', 'append']:
                return self.update.dataframe(df.uniqueId, name, shared_status, column_hash, update_date)
            else:
                raise APIError(f"ID {unique_id} is already registered as DataFrame {df.id}")
        return self.create.dataframe(unique_id, name, shared_status, column_hash, update_date)

    @deprecated_without_replacement('v1.0')
    def feature(self,
                data_source_id: int,
                column_name: str,
                display_name: str,
                data_type: str,
                description: Optional[str] = None,
                granularity: Optional[str] = None,
                status: Optional[str] = None,
                tags: Optional[List[str]] = None,
                git_repo: Optional[str] = None,
                if_exists: str = 'return') -> Feature:
        """
        Creates or updates a feature depending on existence of the defined parameters

        if_exists:  return - returns the source without operating on it
                    edit - edits the existing source
                    new - creates a new source
        """
        column_name = column_name or display_name
        description = description or f"Feature that contains {display_name} data"
        status = status or "Sandboxed"

        ft = self.match.feature(column_name=column_name, data_source_id=data_source_id)
        if ft:
            if if_exists == 'return':
                return ft
            elif if_exists in ['edit', 'replace', 'append']:
                return self.update.feature(id=ft.id, display_name=display_name, column_name=column_name, description=description, status=status, tags=tags or [], git_repo=git_repo)
            else:
                raise APIError(
                    f"Column {column_name} is already registered in DataSource {data_source_id} as Feature {ft.id}")
        return self.create.feature(data_source_id=data_source_id, display_name=display_name, column_name=column_name, description=description, status=status, git_repo=git_repo, tags=tags or [])

    @deprecated_without_replacement('v1.0')
    def features_dict(self, features_dict: dict,
                      trigger_stats: bool = True) -> DataSource:
        """
        Creates or updates features based on values in a dict
        """
        if not ingestion._confirm_valid_dict(dict_in=features_dict):
            raise APIError("Not a valid dict")
        source_table = features_dict.get("sourceTable")
        ds_name = features_dict.get("name", source_table)
        # Using a set here to de-dup tags
        tags = set()
        if isinstance(features_dict.get("tags"), list):
            tags.update(features_dict.get("tags"))
        attributes = []
        # Handle two attribute formats:
        # 1) List of dicts
        if isinstance(features_dict.get("attributes"), list):
            for a in features_dict.get("attributes"):
                if isinstance(a, dict):
                    for k, v in a.items():
                        attributes.append({k: v})
        # 2) dict
        if isinstance(features_dict.get("attributes"), dict):
            for k, v in features_dict.get("attributes").items():
                attributes.append({k: v})
        source_code = features_dict.get("sourceCode")
        source_type = features_dict.get("sourceType")

        # Send a save request here, because it ignores missing fields
        # We don't want to overwrite attributes that aren't present in the dict
        data_source = self.data_source(
            table=source_table,
            name=ds_name,
            source_type=source_type,
            source_code=source_code,
            if_exists="edit")

        columns = data_source.columns

        # Assemble dimensions
        # for col in columns:
        put_dimensions = []
        for dim in features_dict["dimensions"]:
            column_name = dim.get("columnName")
            display_name = dim.get("displayName")
            data_type = dim.get("dataType")
            ds_column = [col.id for col in columns if col.name.upper() == column_name.upper()]
            if len(ds_column) > 0:
                ds_col_id = ds_column[0]
            else:
                APIWarning(f'Potential data issue encountered while publishing dimension: '
                           f'Column {column_name} not found in table {source_table}')
                ds_col_id = None
            # if we get an enum, convert it to str so pydantic doesn't get mad
            if isinstance(data_type, schemas.DataType):
                data_type = data_type.value
            dim_granularity = dim.get("granularity")
            put_dimensions.append(
                schemas.DimensionColumnPut(
                    dataSourceColumnId=ds_col_id,
                    columnName=column_name,
                    displayName=display_name,
                    granularityName=dim_granularity
                )
            )

        # Assemble features
        put_features = []
        for feature in features_dict["features"]:
            display_name = feature.get("displayName")
            column_name = feature.get("columnName")
            data_type = feature.get("dataType")
            ds_column = [col.id for col in columns if col.name.upper() == column_name.upper()]
            if len(ds_column) > 0:
                ds_col_id = ds_column[0]
            else:
                APIWarning(f'Potential data issue encountered while publishing feature: '
                           f'Column {column_name} not found in table {source_table}')
                ds_col_id = None
            # if we get an enum, convert it to str so pydantic doesn't get mad
            if isinstance(data_type, schemas.DataType):
                data_type = data_type.value
            description = feature.get("description", f"Feature that contains {display_name} data")
            # apply source tags to all features...
            feature_tags = tags
            # ...and add feature-specific tags
            if isinstance(feature.get("tags"), list):
                feature_tags.update(feature.get("tags"))
            feature_attributes = attributes
            if isinstance(feature.get("attributes"), list):
                for a in feature.get("attributes"):
                    if isinstance(a, dict):
                        for k, v in a.items():
                            feature_attributes.append({k: v})
            if isinstance(feature.get("attributes"), dict):
                for k, v in feature.get("attributes").items():
                    feature_attributes.append({k: v})
            attributes_put = []
            for kv in feature_attributes:
                for i in kv:
                    attributes_put.append(schemas.Attribute(key=i, value=kv[i]))
            status = "Sandboxed" if features_dict.get("status") == "Sandboxed" else "Productionized"
            put_features.append(
                schemas.FeatureColumnPut(
                    dataSourceColumnId=ds_col_id,
                    columnName=column_name,
                    displayName=display_name,
                    description=description,
                    status=status,
                    tags=list(feature_tags),
                    attributes=attributes_put
                )
            )

        source_out = self.source_features(
            id=data_source.id,
            table=data_source.table,
            name=data_source.name,
            database=data_source.tableDatabase,
            schema=data_source.tableSchema,
            source_code=data_source.sourceCode,
            domain=data_source.domain,
            source_type=data_source.sourceType,
            parent_source_id=data_source.parentId,
            columns=data_source.columns,
            features=put_features,
            dimensions=put_dimensions)

        # Post stats for features
        if trigger_stats:
            self.create.data_source_feature_stats(source_out.id)
        return source_out

    @deprecated_without_replacement('v1.0')
    def column_importance_stats(self, id: str, payload: schemas.feature.FeatureImportanceStats):
        """
        Sends a json payload of importance from a dataFrame to the API so it can render in the WebApp
        """
        # First update timestamp on the DF object
        self.dataframe(unique_id=id, update_date=payload.timestamp, if_exists='edit')
        return self.create.column_importance_stats(id, payload)

    @deprecated_without_replacement('v1.0')
    def feature_importance_stats(self, id: int, payload: schemas.feature.FeatureImportanceStats):
        """
        sends a JSON payload of importance from a collection to the API so it can store and render stats later
        """
        return self.create.feature_importance_stats(id=id, payload=payload)

    @deprecated_without_replacement('v1.0')
    def dataframe_profile(self, id: str, payload: schemas.feature.ColumnProfiles):
        """
        Send a json payload of a dataframe profile so it can render in the WebApp
        """
        # First update timestamp on the DF object
        self.dataframe(unique_id=id, update_date=payload.timestamp, if_exists='edit')
        return self.create.dataframe_profile(id, payload)

    @deprecated_without_replacement('v1.0')
    def source_features(self, id: int,
                        table: str,
                        name: str,
                        database: Optional[str],
                        schema: Optional[str],
                        source_code: Optional[str],
                        domain: Optional[str],
                        source_type: Optional[str],
                        parent_source_id: Optional[int],
                        columns: List[schemas.DataSourceColumnPut],
                        features: Optional[List[schemas.FeatureColumnPut]],
                        dimensions: Optional[List[schemas.DimensionColumnPut]],
                        if_exists: str = 'placebo'
                        ) -> DataSource:
        """
        Creates or updates a DataSource + its Features and Dimensions using the defined parameters
        """
        organization_id = self.api._profile.get('organizationId')
        data_source = schemas.DataSourcePut(name=name,
                                            table=table,
                                            tableDatabase=database,
                                            tableSchema=schema,
                                            sourceCode=source_code,
                                            domain=domain,
                                            sourceType=source_type,
                                            parentId=parent_source_id,
                                            columns=columns,
                                            features=features,
                                            dimensions=dimensions,
                                            organizationId=organization_id)
        response = self.api._put(f"/data-source/{id}/features", data_source.dict(), api_version=1).json()
        return DataSource(api_object=response)




from pydantic import BaseModel
from typing import Any, List, Optional

# Schema to be used for YML files
# TODO: sync with new schema when customers' legacy files are converted
class DimensionYML(BaseModel):
    columnName: str
    dataType: str
    granularity: str
    class Config:
        allow_population_by_field_name = True

class FeatureYML(BaseModel):
    columnName: str
    dataType: str
    description: Optional[str]
    displayName: Optional[str]
    status: Optional[str]
    tags: Optional[List[str]]
    attributes: Optional[dict]
    class Config:
        allow_population_by_field_name = True

class FeaturesYML(BaseModel):
    name: Optional[str] = ''
    sourceTable: str
    dimensions: Optional[List[DimensionYML]]
    features: Optional[List[FeatureYML]]
    sourceCode: Optional[str]
    sourceType: Optional[str]
    class Config:
        allow_population_by_field_name = True


from datetime import datetime
from pydantic import BaseModel, Field
from typing import List, Optional, Dict

from pyrasgo.schemas import dw_table as dw_table_schemas
from pyrasgo.schemas import dataset_column as dataset_column_schemas
from pyrasgo.schemas.snapshot import Snapshot


class Dataset(BaseModel):
    """
    Contract to return from get by id endpoints
    """
    id: int
    name: Optional[str]
    resource_key: str = Field(alias='resourceKey')
    version: int
    description: Optional[str]
    status: Optional[str]

    owner_id: Optional[int] = Field(alias='ownerId')
    organization_id: int = Field(alias='organizationId')
    dw_table_id: Optional[int] = Field(alias='dwTableId')
    dw_operation_set_id: Optional[int] = Field(alias='dwOperationSetId')

    columns: Optional[List[dataset_column_schemas.DatasetColumn]]
    dw_table: Optional[dw_table_schemas.DataTableWithColumns] = Field(alias='dataTable')
    consumer_count: int = Field(alias='consumerCount')

    attributes: Optional[Dict[str, str]]
    tags: Optional[List[str]]
    snapshots: Optional[List[Snapshot]]


    create_timestamp: datetime = Field(alias='createTimestamp')
    create_author: int = Field(alias='createdBy')
    update_timestamp: datetime = Field(alias='updateTimestamp')
    update_author: int = Field(alias='updatedBy')

    class Config:
        allow_population_by_field_name = True


class DatasetBulk(BaseModel):
    """
    Contract to return from get list endpoints
    """
    id: int
    name: Optional[str]
    version: int
    resource_key: str = Field(alias='resourceKey')
    description: Optional[str]

    owner_id: Optional[int] = Field(alias='ownerId')
    organization_id: int = Field(alias='organizationId')

    dw_table_id: Optional[int] = Field(alias='dwTableId')
    dw_table: Optional[dw_table_schemas.DataTable] = Field(alias='dataTable')

    dw_operation_set_id: Optional[int] = Field(alias='dwOperationSetId')
    dataset_dependencies: List[int] = Field(alias='datasetDependencies')

    column_count: int = Field(alias='columnCount')
    consumer_count: int = Field(alias='consumerCount')

    attributes: Optional[Dict[str, str]]
    tags: Optional[List[str]]

    create_timestamp: datetime = Field(alias='createTimestamp')
    create_author: int = Field(alias='createdBy')
    update_timestamp: datetime = Field(alias='updateTimestamp')
    update_author: int = Field(alias='updatedBy')

    class Config:
        allow_population_by_field_name = True


class DatasetCreate(BaseModel):
    """
    Contract to accept in post endpoints
    """
    name: str
    resource_key: Optional[str] = Field(alias='resourceKey')
    description: Optional[str]
    status: Optional[str]
    dw_table_id: Optional[int] = Field(alias='dwTableId')
    dw_operation_set_id: Optional[int] = Field(alias='dwOperationSetId')
    attributes: Optional[Dict[str, str]]
    auto_generate_stats: bool = Field(alias='autoGenerateStats', default=True)

    class Config:
        allow_population_by_field_name = True


class DatasetUpdate(BaseModel):
    """
    Contract to accept in put endpoints
    """
    name: Optional[str]
    resource_key: Optional[str] = Field(alias='resourceKey')
    description: Optional[str]
    status: Optional[str]
    owner_id: Optional[int] = Field(alias='ownerId')
    dw_table_id: Optional[int] = Field(alias='dwTableId')
    dataset_columns: Optional[List[dataset_column_schemas.DatasetColumnUpdateById]]
    attributes: Optional[dict]

    class Config:
        allow_population_by_field_name = True



import pandas as pd
from typing import Dict, List, Optional

from pyrasgo.utils.versioning import deprecated_without_replacement

from .error import APIError
from pyrasgo import config
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, DataSource, Dataset


class Read():

    def __init__(self):
        from . import Get
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()
        self.get = Get()

    @deprecated_without_replacement('v1.0')
    def collection_data(self,
                        id: int,
                        filters: Optional[Dict[str, str]] = None,
                        limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Collection

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        collection = self.get.collection(id)
        if collection:
            try:
                table_metadata = collection._make_table_metadata()
                query, values = self.data_warehouse.make_select_statement_filter_dict(table_metadata, filters, limit)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"Collection table is not reachable: {e}")
        raise APIError("Collection does not exist")

    @deprecated_without_replacement('v1.0')
    def feature_data(self,
                     id: int,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Feature data

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        feature = self.get.feature(id)
        if feature.sourceTable:
            try:
                table_metadata = feature._make_table_metadata()
                #TODO: if we ever support multiple features, add them to this line -
                features = feature.columnName
                indices = ','.join(feature.indexFields)
                columns = indices +', '+features
                query, values = self.data_warehouse.make_select_statement_filter_dict(table_metadata, filters, limit, columns)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"Feature table is not reachable: {e}")
        raise APIError("Feature table does not exist")

    @deprecated_without_replacement('v1.0')
    def source_data(self,
                    id: int,
                    filters: Optional[Dict[str, str]] = None,
                    limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo DataSource

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        data_source = self.get.data_source(id)
        if data_source:
            try:
                table_metadata = data_source._make_table_metadata()
                query, values = self.data_warehouse.make_select_statement_filter_dict(table_metadata, filters, limit)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"DataSource table is not reachable: {e}")
        raise APIError("DataSource does not exist")

    def dataset(
            self,
            id: Optional[int] = None,
            dataset: Optional[Dataset] = None,
            filters: Optional[List[str]] = None,
            order_by: Optional[List[str]] = None,
            columns: Optional[List[str]] = None,
            limit: Optional[int] = None,
            snapshot_index: Optional[int] = None
    ) -> pd.DataFrame:
        """
        Constructs and returns pandas DataFrame from the specified Rasgo Dataset

        You can supply SQL WHERE clause filters, order the dataset by columns, only
        return selected columns, and add a return limit as well

        Example:
            ```
            rasgo.read.dataset(
                id=74,
                filters=['SALESTERRITORYKEY = 1', 'TOTALPRODUCTCOST BETWEEN 1000 AND 2000'],
                order_by=['TOTALPRODUCTCOST'],
                columns=['PRODUCTKEY', 'TOTALPRODUCTCOST', 'SALESTERRITORYKEY'],
                limit=50
            )
            ```

        Args:
            id: dataset id to read into df
            dataset: Dataset obj to read into df
            filters: List of SQL WHERE filters strings to filter on returned df
            order_by: List of columns to order by in returned dataset
            columns: List of columns to return in the df
            limit: Only return this many rows in the df
            snapshot: the index of a snapshot from Dataset.snapshots to read 
        """
        # Validate one dataset passed in, id or dataset obj
        if not dataset and id is None:
            raise ValueError(f"Must pass either a valid dataset ID or Dataset "
                             f"object to read into a DataFrame")
        if not dataset:
            # Note: Func below already raises API error if dataset with id doesn't exist
            dataset = self.get.dataset(id)

        # Require the operation set on the DS to make
        # sure table is created before reading
        dataset._require_operation_set()

        # Get the FQTN from snapshot or current dataset
        if snapshot_index:
            try:    
                fqtn = dataset._api_dataset.snapshots[snapshot_index].fqtn
            except IndexError:
                raise ValueError(f"Snapshot index does not exist")
        else:
            fqtn = dataset.fqtn

        if dataset:
            try:
                query = self.data_warehouse.make_select_statement(
                    table_metadata={'fqtn': fqtn},
                    filters=filters,
                    order_by=order_by,
                    columns=columns,
                    limit=limit
                )
                return self.data_warehouse.query_into_dataframe(query)
            except Exception as e:
                raise APIError(f"Dataset table is not reachable: {e}")



import json
from typing import Dict
from urllib.parse import quote, urlencode

import requests
from pyrasgo.api.error import APIError
from requests import Response

from .session import SessionMeta, generate_headers


class Connection(object, metaclass=SessionMeta):
    """
    Base class for all Rasgo objects to facilitate API calls.
    """
    def __init__(self, api_key=None):
        if api_key:
            self._api_key = api_key

    def _get_profile(self):
        return self._profile

    def _get_default_namespace(self) -> Dict:
        org = self._profile.get('organization')
        return {
            'database': org.get('database'),
            'schema': org.get('schema')
        }

    def _url(self, resource, api_version=None):
        if '/' == resource[0]:
            resource = resource[1:]
        protocol = 'http' if self._environment.value == 'localhost' else 'https'
        return f"{protocol}://{self._environment.value}/{'' if api_version is None else f'v{api_version}/'}{resource}"

    def find(self, resource, equality_filters):
        search_strings = []
        for k, v in equality_filters.items():
            filterstr = k + "||$eq||" + v
            search_strings.append(filterstr)
        params = {
            "filter": search_strings
        }
        params = urlencode(params, quote_via=quote, doseq=True)
        response = self._get(resource, params)
        if response.status_code == 404:
            return None
        else:
            self._raise_internal_api_error_if_any(response)
            return response.json()

    def _delete(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs DELETE request to Rasgo API as defined within the class instance.

        :param resource: Target resource to DELETE from API.
        :param _json: JSON object to send in DELETE request
        :param params: Additional parameters to specify for POST request.
        :return: Response object containing content returned.
        """
        response = requests.delete(self._url(resource, api_version),
                                   json=_json,
                                   headers=generate_headers(self._api_key),
                                   params=params or {})
        self._raise_internal_api_error_if_any(response)
        return response

    def _get(self, resource, params=None, api_version=None) -> requests.Response:
        """
        Performs GET request to Rasgo API as defined within the class instance.

        :param endpoint: Target resource to GET from API.
        :param params: Additional parameters to specify for GET request.
        :return: Response object containing content returned.
        """
        response = requests.get(self._url(resource, api_version),
                                headers=generate_headers(self._api_key),
                                params=params or {})
        self._raise_internal_api_error_if_any(response)
        return response

    def _patch(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs PATCH request to Rasgo API as defined within the class instance.

        :param resource: Target resource to PATCH from API.
        :param _json: JSON object to send in PATCH request
        :param params: Additional parameters to specify for PATCH request.
        :return: Response object containing content returned.
        """
        response = requests.patch(self._url(resource, api_version),
                                  json=_json,
                                  headers=generate_headers(self._api_key),
                                  params=params or {})
        self._raise_internal_api_error_if_any(response)
        return response

    def _post(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs POST request to Rasgo API as defined within the class instance.

        :param resource: Target resource to POST from API.
        :param _json: JSON object to send in POST request
        :param params: Additional parameters to specify for POST request.
        :return: Response object containing content returned.
        """
        response = requests.post(self._url(resource, api_version),
                                 json=_json,
                                 headers=generate_headers(self._api_key),
                                 params=params or {})
        self._raise_internal_api_error_if_any(response)
        return response

    def _put(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs PUT request to Rasgo API as defined within the class instance.

        :param resource: Target resource to PUT from API.
        :param _json: JSON object to send in PUT request
        :param params: Additional parameters to specify for PUT request.
        :return: Response object containing content returned.
        """
        response = requests.put(self._url(resource, api_version),
                                json=_json,
                                headers=generate_headers(self._api_key),
                                params=params or {})
        self._raise_internal_api_error_if_any(response)
        return response

    def _get_items(self, resource, equality_filters):
        search_strings = []
        for k, v in equality_filters.items():
            filterstr = k + "||$eq||" + v
            search_strings.append(filterstr)
        params = {
            "filter": search_strings
        }
        params = urlencode(params, quote_via=quote, doseq=True)
        try:
            response = self._get(resource, params)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return None
            else:
                raise e
        if response.status_code == 404:
            return None
        else:
            self._raise_internal_api_error_if_any(response)
            return response.json()

    def _get_item(self, resource, equality_filters, allow_multiple=False):
        results = self._get_items(resource, equality_filters)
        if results is None:
            return None
        elif 1 < len(results):
            if allow_multiple:
                return results[-1]
            else:
                raise ValueError("Multiple returned for {}, check your filters".format(resource))
        elif 0 == len(results):
            return None
        else:
            return results[0]

    def _raise_internal_api_error_if_any(self, response: Response) -> None:
        """
        Raise an internal API error if any occurred

        Show proper APIError Message in PyRasgo by parsing
        FastAPI HttpError msg
        """
        try:
            response.raise_for_status()
        except:
            try:
                err_details = response.json()
            except json.JSONDecodeError:
                err_details = response.text
            raise APIError(
                f"Internal API Error when making {response.request.method} request\n\n"
                f"Status Code: {response.status_code}\n\n"
                f"Internal Error Details: {err_details}"
            )



import pandas as pd
from typing import List, Optional
import inspect

from pyrasgo.api.error import APIError
from pyrasgo.utils import track_experiment

class Prune():

    def __init__(self, experiment_id = None):
        from pyrasgo.storage import Evaluate

        self._experiment_id = experiment_id
        self.evaluate = Evaluate(experiment_id=experiment_id)

    @track_experiment
    def columns_with_missing_data(self, df: pd.DataFrame,
                                  columns: List[str] = None,
                                  threshold: float = 0) -> pd.DataFrame:
        """
        Returns a copy of your dataframe after removing columns
        containing NULL or NaN values exceeding a threshold count

        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            columns: (Optional) List[str]:
                List of column names to check for null, Defaults to All
            threshold: (Optional) float:
                Null percentage threshold above which a column will be dropped, Defaults to 0%

        Returns
        -------
            pandas DataFrame
        """
        df_out = df.copy(deep=True)
        column_with_nan = df_out.columns[df_out.isnull().any()]
        for column in column_with_nan:
            if columns:
                if column in columns:
                    if df_out[column].isnull().sum()*100.0/df_out.shape[0] >= threshold:
                        df_out = df_out.drop(column, 1)
                        print(f'Column deleted: {column}')
            else:
                if df_out[column].isnull().sum()*100.0/df_out.shape[0] >= threshold:
                    df_out = df_out.drop(column, 1)
                    print(f'Column deleted: {column}')
        return df_out

    @track_experiment
    def duplicate_rows(self, df: pd.DataFrame,
                       columns: List[str] = None
                       ) -> pd.DataFrame:
        """
        Returns a copy of your DataFrame with duplicate rows removed

        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            columns: (Optional) List[str]:
                List of column names to check for duplicates in

        Returns
        -------
            pandas DataFrame
        """
        df_out = df.copy(deep=True)
        #df_out = df_out.reset_index(drop=False)
        #df_out.rename(columns = {'index':'indexInOrigDF'}, inplace = True)
        if columns:
            drop_us = pd.Int64Index([])
            for column in columns:
                drop_us = drop_us.append(df_out[df_out.duplicated([column])].index)
        else:
            drop_us = df_out.index[df_out.duplicated()]
        print(f'Dropping {drop_us.shape[0]} rows')
        df_out = pd.DataFrame(df_out.drop(drop_us, 0))
        return df_out

    @track_experiment
    def features(self, df: pd.DataFrame,
                 target_column: str,
                 timeseries_index: str = None,
                 exclude_columns: List[str] = None,
                 top_n: int = None,
                 top_n_pct: float = None,
                 pct_of_top_feature: float = None
                 ) -> pd.DataFrame:
        """
        Drops columns from a dataframe that are below a threshold of feature importance

        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            target_columns: str:
                Column name of target feature
            timeseries_index: (Optional) str:
                Name of column to use as timeseries index.
                If passed, rows will be ordered by column and split.
                If None is passed, rows will be split randomly.
            top_n: int:
                The number of features to keep, ordered by feature importance
            top_n_pct: float:
                The percentage of features to keep, ordered by feature importance
            pct_of_top_feature: float:
                The percentage of importance of the most import feature, any columns with importance below this percentage will be dropped
            exclude_columns: List[str]:
                Column names of features to be filered out from calculation

        Returns
        -------
            pandas DataFrame
        """
        # Runtime checks
        if top_n is not None:
            method = 'TOPN'
            print(f'Prune Method: Keeping top {top_n} features')
        elif top_n_pct is not None:
            method = 'TOPPCT'
            print(f'Prune Method: Keeping top {top_n_pct} of features')
        elif pct_of_top_feature is not None:
            method = 'PCTOF'
            print(f'Prune Method: Keeping features with importance above {pct_of_top_feature} pct of max feature')
        else:
            raise APIError('Must pass in one of the following parameters: top_n=#, top_n_pct=.#, pct_of_top_feature=.#')

        # Copy df so we can alter it without impacting source data
        p_df = df.copy(deep=True)

        # Calculate feature_importance and return the result as a dict
        fi_dict = self.evaluate.feature_importance(df=p_df, target_column=target_column, timeseries_index=timeseries_index, exclude_columns=exclude_columns, return_cli_only=True)
        fi_s = pd.Series(fi_dict.get("featureImportance"))
        fi_df = pd.DataFrame({'Feature':fi_s.index, 'Importance':fi_s.values})
        fi_df.sort_values(by='Importance', ascending=False, inplace=True)

        if method == 'TOPN':
            row_threshold = top_n-1
            low_importance_features = fi_df.iloc[row_threshold:]['Feature']
            print(f'Dropped features not in top {top_n}:', low_importance_features.to_list())
        if method == 'TOPPCT':
            row_threshold = round(fi_df.shape[0] * top_n_pct)
            low_importance_features = fi_df.iloc[row_threshold:]['Feature']
            print(f'Dropped features not in top {top_n_pct} pct:', low_importance_features.to_list())
        if method == 'PCTOF':
            max_fi = fi_df['Importance'].max()
            importance_cutoff = max_fi * pct_of_top_feature
            low_importance_features = fi_df[fi_df['Importance'] <= importance_cutoff]['Feature']
            print(f'Dropped features below importance threshold {importance_cutoff}:', low_importance_features.to_list())

        p_df = p_df.drop(columns=low_importance_features)
        return p_df

    @track_experiment
    def rows_with_missing_data(self, df: pd.DataFrame,
                               columns: List[str] = None
                               ) -> pd.DataFrame:
        """
        Returns a copy of your DataFrame after removing all rows
        containing NULL or NaN values in any of their columns

        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            columns: (Optional) List[str]:
                List of column names to check for nulls, Defaults to All

        Returns
        -------
            pandas DataFrame
        """
        df_out = df.copy(deep=True)
        if columns:
            index_with_nan = pd.Int64Index([])
            for column in columns:
                index_with_nan = index_with_nan.append(df_out[df_out[column].isnull()].index)
        else:
            index_with_nan = df_out.index[df_out.isnull().any(axis=1)]
        print(f'Dropping {index_with_nan.shape[0]} rows')
        df_out = pd.DataFrame(df_out.drop(index_with_nan))
        return df_out



import time
from tqdm import tqdm
from typing import Dict, List, Optional, Union

from .feature import Feature, FeatureList

import pandas as pd
from pyrasgo.api.connection import Connection
from pyrasgo.api.error import APIError
from pyrasgo.schemas import collection as schema
from pyrasgo.schemas.attributes import Attribute, CollectionAttributeBulkCreate
from pyrasgo.schemas.enums import ModelType


class Collection(Connection):
    """
    Stores a Rasgo Collection
    """
    def __init__(self, api_object, **kwargs):
        super().__init__(**kwargs)
        self._fields = schema.Collection(**api_object)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Collection(id={self.id}, name={self.name}, type={self.modelType.name}, " \
               f"authorId={self.authorId}, isShared={self.isShared})"

    def __getattr__(self, item):
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            self._refresh()
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")


# ----------
# Properties
# ----------

    @property
    def attributes(self) -> dict:
        """
        Helper property to convert a list of dicts into a single dict
        """
        attr_dict = {}
        response = self._get(f"/models/{self.id}/attributes", api_version=1).json()
        for a in response:
            attr_dict.update({a['key']: a['value']})
        return attr_dict

    @property
    def author(self) -> str:
        """
        Returns the author object of the Collection (includes credentials if user is author)
        """
        if self._profile['id'] == self.authorId:
            author = self._profile
        else:
            author = self._get(f"/users/{self.authorId}", api_version=1).json()
        return f'Author({author["id"]}, {author["firstName"]} {author["lastName"]})'

    @property
    def dimensions(self):
        if not self._fields.dimensions:
            self._refresh()
        return self._fields.dimensions

    @property
    def features(self) -> FeatureList:
        """
        Returns the Features within the Collection
        """
        try:
            return FeatureList([feature.dict() for feature in self._fields.features])
        except (AttributeError, TypeError):
            self._refresh()
        return FeatureList([feature.dict() for feature in self._fields.features])

    @features.setter
    def features(self, features: Union[List[Feature], FeatureList, Feature]):
        # Update the namespace
        all_features = [Feature(api_object=feature.dict()) for feature in self.features + features]
        self._fields.__setattr__('features', all_features)

    @property
    def modelType(self) -> ModelType:
        """
        Returns the type of the Collection
        """
        return ModelType(self._fields.type)

# -------
# Methods
# -------

    def get_compatible_features(self, return_dict: bool = False) -> Union[Dict, pd.DataFrame]:
        """
        Return all the compatible features which could be added to this collection

        Args:
            return_dict: Set to True to return as Dict instead of df
        """
        try:
            url = f"/models/{self.id}/compatible_features"
            all_compatible_features = self._get(url, api_version=1).json()
            if return_dict:
                return all_compatible_features
            return pd.DataFrame.from_dict(all_compatible_features)
        except:
            raise APIError(f"Collection {self.id} does not exist or this "
                           f"API key does not have access to it.")

    def preview(self, limit: int = 10) -> pd.DataFrame:
        """
        Preview the first x rows of this collection rows as a df

        Args:
            limit: integer limit for number of rows returned (Default 10)

        Returns:
            Pandas Dataframe Preview of Dataframe
        """
        return self.read_into_df(limit=limit)

    def add_attributes(self, attributes: List[dict]) -> List[dict]:
        if not isinstance(attributes, list):
            raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: '
                             '[{"key": "value"}, {"key": "value"}]')
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: '
                                 '[{"key": "value"}, {"key": "value"}]')
            for k, v in kv.items():
                attr.append(Attribute(key=k, value=v))
        attr_in = CollectionAttributeBulkCreate(collectionId=self.id, attributes=attr)
        return self._put(f"/models/{self.id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()

    def add_feature(self, feature: Feature) -> None:
        """
        Adds a single feature to the Collection
        """
        self.features = feature
        self._patch(f"/models/{self.id}/features", api_version=1,
                    _json={"featureIds": [str(feature.id)]})

    def add_features(self, features: Union[List[Feature], FeatureList]) -> None:
        """
        Adds a FeatureList or a list of Features to the Collection
        """
        self.features = features
        self._patch(f"/models/{self.id}/features", api_version=1,
                    _json={"featureIds": [str(feature.id) for feature in features]})

    def generate_training_data(self, trigger_stats: bool = True) -> None:
        """
        Triggers the generation of the Collection's training data.
        """
        self._post(f"/trainings/",
                   _json={"modelId": int(self.id),
                          "userId": self._profile['id']},
                   params={"trigger_stats": trigger_stats} if trigger_stats is not None else {},
                   api_version=0)
        for _ in tqdm(range(0, 10), leave=False):
            if self.is_data_ready():
                break
            time.sleep(1)

    def is_data_ready(self) -> bool:
        """
        Performs check against API for training data readiness, if true, a dataframe can be pulled down.
        :return:
        """
        r = self._get("/trainings/latest", api_version=0)
        if any([model['state'] == 'done' for model in r.json() if model['model']['id'] == self.id]):
            return True
        if any([model['state'] == 'error' for model in r.json() if model['model']['id'] == self.id]):
            raise SystemError("There's been an issue with the training data generation")
        return False

    def read_into_df(self,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        from pyrasgo.api.read import Read
        return Read().collection_data(id=self.id, filters=filters, limit=limit)

    def _refresh(self) -> None:
        """
        Updates the Collection's attributes from the API
        """
        this_collection_dict = self._get(f"/models/{self.id}", api_version=1).json()
        self._fields = schema.Collection(**this_collection_dict)

    def rename(self, new_name: str) -> None:
        """
        Updates a Collection's display name
        """
        print(f"Renaming Collection {self.id} from {self.name} to {new_name}")
        collection = schema.CollectionUpdate(id=self.id, name=new_name)
        self._fields = schema.Collection(**self._patch(
            f"/models/{self.id}/details",
            api_version=1,
            _json=collection.dict(exclude_unset=True, exclude_none=True)
        ).json())

    def render_sql_definition(self) -> str:
        sql_str = self._get(f"/models/{self.id}/sql", api_version=1).json()
        return sql_str

    def share(self, share: bool = True) -> bool:
        """
        Share the Collection with your organization
        :param share: True to make your Collection public. False to make your Collection private
        """
        response = self._patch(f"/models/{self.id}/details", api_version=1, _json={"isShared": share}).json()
        return response['isShared']

    def toggle_stats(self, trigger_stats: bool) -> dict:
        """
        Determine whether to run correlation and join stats when training the Collection
        """
        response = self.add_attributes([{"_trigger_stats": trigger_stats}])
        for r in response:
            if r.get("key") == "_trigger_stats":
                return {r["key"]: r["value"]}

    def _make_table_metadata(self) -> dict:
        organization = self._get_profile().get("organization")
        metadata = {
            "database": organization.get("database"),
            "schema": organization.get("schema"),
            "table": self._fields.dataTableName,
        }
        return metadata



from types import FunctionType

from collections import OrderedDict
import logging
import os
import yaml

from pyrasgo.primitives import DataSource
from pyrasgo.schemas import DataType, FeaturesYML


RASGO_DICT: dict = {
"name": "optional name for DataSource",
"sourceTable": "mandatory DB.SCHEMA.TABLE features are stored in",
"features": [
                {
                    "columnName": "mandatory_field_name_in_table",
                    "displayName": "Optional Pretty Name",
                    "dataType": "mandatory sql type",
                    "description": "optional display text",
                    "status": "Productionized | Sandboxed",
                    "tags": [
                                "optional",
                                "list of strings",
                                "apply only to this feature"
                            ],
                    "attributes":  [
                                        {"key": "value"},
                                        {"optional": "apply only to this features"}
                                   ]
                }
            ],
"dimensions": [
                {
                    "columnName": "mandatory_field_name_in_table",
                    "displayName": "Optional Pretty Name",
                    "dataType": "mandatory sql type",
                    "granularity": "mandatory noun ..."
                }
            ],
"tags": [
            "optional",
            "list of strings",
            "apply to all features"
        ],
"attributes": [
                {"key": "value"},
                {"optional": "apply to all features"}
              ],
"sourceCode": "Optional valid SQL or python string",
"sourceType": "Table | CSV | Dataframe"
}

def load_features_from_yaml(*, file_name: str, directory: str = None) -> DataSource:
    if directory is None:
        directory = os.getcwd()

    if os.path.splitext(f"{directory}/{file_name}")[1] in ['yaml', 'yml']:
        raise ValueError("Must provide valid yaml file")

    with open(f"{directory}/{file_name}") as _yaml:
        return DataSource(api_object=yaml.load(_yaml, Loader=yaml.SafeLoader))

def load_features_from_dict(*, features_dict: dict) -> DataSource:
    return DataSource(api_object=features_dict)

def save_features_to_yaml(data_source: FeaturesYML, *,
                          file_name: str,
                          directory: str = None,
                          overwrite: bool = True) -> None:
    if directory is None:
        directory = os.getcwd()

    if directory[-1] == "/":
        directory = directory[:-1]

    if file_name.split(".")[-1] not in ['yaml', 'yml']:
        file_name += ".yaml"

    if os.path.exists(f"{directory}/{file_name}") and overwrite:
        logging.warning(f"Overwriting existing file {file_name} in directory: {directory}")

    safe_dumper = yaml.SafeDumper
    safe_dumper.add_representer(DataType, lambda self, data: self.represent_str(str(data.value)))
    safe_dumper.add_representer(OrderedDict, lambda self, data: self.represent_mapping('tag:yaml.org,2002:map',
                                                                                       data.items()))
    safe_dumper.ignore_aliases = lambda self, data: True

    with open(f"{directory}/{file_name}", "w") as _yaml:
        yaml.dump(data=OrderedDict(data_source.dict(exclude_unset=True, by_alias=True)), Dumper=safe_dumper, stream=_yaml)

def save_features_to_dict(data_source: DataSource, overwrite: bool = True) -> dict:
    return FeaturesYML(
        name = data_source.name,
        sourceTable = data_source.dataTable.fqtn,
        dimensions = [{"columnName": d.columnName, "dataType": d.dataType, "granularity": d.granularity.name} for d in data_source.dimensions],
        features = data_source.features,
        sourceCode = data_source.sourceCode,
        sourceType = data_source.sourceType
    ).dict(exclude_unset=False, by_alias=True)

def _confirm_valid_dict(dict_in: dict, version: int = None) -> bool:
    if not dict_in.get("sourceTable"):
        print("Missing table")
        return False
    if not dict_in.get("features"):
        print("Missing features")
        return False
    for f in dict_in.get("features"):
        if not f.get("columnName") or not f.get("dataType"):
            print("Missing feature attributes")
            return False
    if not dict_in.get("dimensions"):
        print("Missing dimensions")
        return False
    for d in dict_in.get("dimensions"):
        if not d.get("columnName") or not d.get("dataType") or not d.get("granularity"):
            print("Missing dimension attributes")
            return False
    return True